import pickle
import random
import tempfile

import pytest

from hstrat.hstrat import stochastic_algo


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_init(replicate):
    random.seed(replicate)
    assert (
        stochastic_algo.Policy().GetSpec()
        == stochastic_algo.Policy(
            policy_spec=stochastic_algo.PolicySpec(),
        ).GetSpec()
    )

    policy = stochastic_algo.Policy()

    # invariants
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBound)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBound)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundPessimalRank)
    assert callable(policy.CalcNumStrataRetainedUpperBound)
    # scrying
    assert policy.CalcMrcaUncertaintyAbsExact is None
    assert policy.CalcMrcaUncertaintyRelExact is None
    assert policy.CalcNumStrataRetainedExact is None
    assert policy.CalcRankAtColumnIndex is None
    assert policy.IterRetainedRanks is None
    # enactment
    assert callable(policy.GenDropRanks)


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_eq(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    assert policy == policy
    assert policy == stochastic_algo.Policy()
    assert policy == policy.WithoutCalcRankAtColumnIndex()
    assert (
        policy.WithoutCalcRankAtColumnIndex()
        == policy.WithoutCalcRankAtColumnIndex()
    )


def test_pickle():
    original = stochastic_algo.Policy()
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_GetSpec(replicate):
    random.seed(replicate)
    assert stochastic_algo.Policy().GetSpec() == stochastic_algo.PolicySpec()


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_WithoutCalcRankAtColumnIndex(replicate):
    random.seed(replicate)

    original = stochastic_algo.Policy()
    stripped = original.WithoutCalcRankAtColumnIndex()

    assert stripped.CalcRankAtColumnIndex is None

    assert (
        original.CalcMrcaUncertaintyAbsUpperBound
        == stripped.CalcMrcaUncertaintyAbsUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBound
        == stripped.CalcMrcaUncertaintyRelUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundPessimalRank
    )
    assert (
        original.CalcNumStrataRetainedUpperBound
        == stripped.CalcNumStrataRetainedUpperBound
    )
    # scrying
    assert (
        original.CalcMrcaUncertaintyAbsExact
        == stripped.CalcMrcaUncertaintyAbsExact
    )
    assert (
        original.CalcMrcaUncertaintyRelExact
        == stripped.CalcMrcaUncertaintyRelExact
    )
    assert (
        original.CalcNumStrataRetainedExact
        == stripped.CalcNumStrataRetainedExact
    )
    assert original.IterRetainedRanks == stripped.IterRetainedRanks
    # enactment
    assert original.GenDropRanks == stripped.GenDropRanks

    # test chaining
    assert stochastic_algo.Policy().WithoutCalcRankAtColumnIndex() == stripped


def test_repr():
    policy = stochastic_algo.Policy()
    assert policy.GetSpec().GetAlgoIdentifier() in repr(policy)


def test_str():
    policy = stochastic_algo.Policy()
    assert policy.GetSpec().GetAlgoTitle() in str(policy)



import itertools as it
import operator
import random

import pytest

from hstrat._auxiliary_lib import find_bounds
from hstrat.hstrat import (
    HereditaryStratigraphicColumn,
    HereditaryStratumOrderedStoreDict,
    stochastic_algo,
)


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_policy_consistency(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    column = HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_ordered_store=HereditaryStratumOrderedStoreDict,
    )
    spec = policy.GetSpec()
    instance = stochastic_algo.CalcMrcaUncertaintyRelUpperBound(spec)
    for num_strata_deposited in range(1, 200):
        for actual_mrca_rank in it.chain(
            range(min(100, num_strata_deposited)),
            range(100, num_strata_deposited, 47),
        ):
            recency = max(
                num_strata_deposited - 1 - actual_mrca_rank,
                1,
            )
            lb, ub = find_bounds(
                query=actual_mrca_rank,
                iterable=column.IterRetainedRanks(),
                initializer=(0, num_strata_deposited - 1),
                filter_above=operator.ge,
                filter_below=operator.le,
            )
            policy_requirement = (ub - lb) / recency
            for which in (
                instance,
                stochastic_algo.CalcMrcaUncertaintyRelUpperBound(spec),
            ):
                assert (
                    which(
                        policy,
                        num_strata_deposited,
                        num_strata_deposited,
                        actual_mrca_rank,
                    )
                    >= policy_requirement
                )
        column.DepositStratum()


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_eq(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    spec = policy.GetSpec()
    instance = stochastic_algo.CalcMrcaUncertaintyRelUpperBound(spec)

    assert instance == instance
    assert instance == stochastic_algo.CalcMrcaUncertaintyRelUpperBound(
        spec,
    )
    assert instance is not None


@pytest.mark.parametrize(
    "replicate",
    range(5),
)
def test_negative_index(replicate):
    random.seed(replicate)
    policy = stochastic_algo.Policy()
    spec = policy.GetSpec()
    instance = stochastic_algo.CalcMrcaUncertaintyRelUpperBound(spec)

    for diff in range(1, 100):
        assert instance(policy, 100, 100, -diff,) == instance(
            policy,
            100,
            100,
            99 - diff,
        )

        assert instance(policy, 101, 100, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 150, 100, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 101, -diff,) == instance(
            policy,
            101,
            100,
            99 - diff,
        )

        assert instance(policy, 100, 150, -diff,) == instance(
            policy,
            150,
            100,
            99 - diff,
        )



import pickle
import tempfile

import pytest

from hstrat.hstrat import recency_proportional_resolution_algo


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_init(recency_proportional_resolution):
    assert (
        recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution
        ).GetSpec()
        == recency_proportional_resolution_algo.Policy(
            policy_spec=recency_proportional_resolution_algo.PolicySpec(
                recency_proportional_resolution
            ),
        ).GetSpec()
    )

    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )

    # invariants
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBound)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBound)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundPessimalRank)
    assert callable(policy.CalcNumStrataRetainedUpperBound)
    # scrying
    assert callable(policy.CalcMrcaUncertaintyAbsExact)
    assert callable(policy.CalcMrcaUncertaintyRelExact)
    assert callable(policy.CalcNumStrataRetainedExact)
    assert callable(policy.CalcRankAtColumnIndex)
    assert callable(policy.IterRetainedRanks)
    # enactment
    assert callable(policy.GenDropRanks)


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_eq(recency_proportional_resolution):
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    assert policy == policy
    assert policy == recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    assert not policy == policy.WithoutCalcRankAtColumnIndex()
    assert (
        policy.WithoutCalcRankAtColumnIndex()
        == policy.WithoutCalcRankAtColumnIndex()
    )
    assert not policy == recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution + 1
    )


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_pickle(recency_proportional_resolution):
    original = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_GetSpec(recency_proportional_resolution):
    assert recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    ).GetSpec() == recency_proportional_resolution_algo.PolicySpec(
        recency_proportional_resolution
    )


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
def test_WithoutCalcRankAtColumnIndex(recency_proportional_resolution):

    original = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    stripped = original.WithoutCalcRankAtColumnIndex()

    assert stripped.CalcRankAtColumnIndex is None

    assert (
        original.CalcMrcaUncertaintyAbsUpperBound
        == stripped.CalcMrcaUncertaintyAbsUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBound
        == stripped.CalcMrcaUncertaintyRelUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundPessimalRank
    )
    assert (
        original.CalcNumStrataRetainedUpperBound
        == stripped.CalcNumStrataRetainedUpperBound
    )
    # scrying
    assert (
        original.CalcMrcaUncertaintyAbsExact
        == stripped.CalcMrcaUncertaintyAbsExact
    )
    assert (
        original.CalcMrcaUncertaintyRelExact
        == stripped.CalcMrcaUncertaintyRelExact
    )
    assert (
        original.CalcNumStrataRetainedExact
        == stripped.CalcNumStrataRetainedExact
    )
    assert original.IterRetainedRanks == stripped.IterRetainedRanks
    # enactment
    assert original.GenDropRanks == stripped.GenDropRanks

    # test chaining
    assert (
        recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution,
        ).WithoutCalcRankAtColumnIndex()
        == stripped
    )


def test_repr():
    recency_proportional_resolution = 1
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    assert str(recency_proportional_resolution) in repr(policy)
    assert policy.GetSpec().GetAlgoIdentifier() in repr(policy)


def test_str():
    recency_proportional_resolution = 1
    policy = recency_proportional_resolution_algo.Policy(
        recency_proportional_resolution
    )
    assert str(recency_proportional_resolution) in str(policy)
    assert policy.GetSpec().GetAlgoTitle() in str(policy)



from hstrat._auxiliary_lib import count_trailing_zeros


def test_no_trailing_zeros():
    assert count_trailing_zeros(0b1) == 0
    assert count_trailing_zeros(-0b1) == 0
    assert count_trailing_zeros(0b11) == 0
    assert count_trailing_zeros(-0b11) == 0


def test_single_trailing_zero():
    assert count_trailing_zeros(0b100110) == 1
    assert count_trailing_zeros(-0b100110) == 1
    assert count_trailing_zeros(0b10) == 1
    assert count_trailing_zeros(-0b10) == 1


def test_multiple_trailing_zeros():
    assert count_trailing_zeros(0b100) == 2
    assert count_trailing_zeros(-0b100) == 2
    assert count_trailing_zeros(0b1100) == 2
    assert count_trailing_zeros(-0b1100) == 2
    assert count_trailing_zeros(0b10100) == 2
    assert count_trailing_zeros(-0b10100) == 2
    assert count_trailing_zeros(0b11100) == 2
    assert count_trailing_zeros(-0b11100) == 2
    assert count_trailing_zeros(0b1000) == 3
    assert count_trailing_zeros(-0b1000) == 3



import numpy as np
import pandas as pd

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_oldest_root import alifestd_mark_oldest_root
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_delete_trunk_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> pd.DataFrame:
    """Delete entries masked by `is_trunk` column.

    Masked entries must be contiguous, meaning that no non-trunk entry can
    be an ancestor of a trunk entry. Children of deleted entries will become
    roots.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_collapse_trunk_asexual
    """
    if "is_trunk" not in phylogeny_df:
        raise ValueError(
            "`is_trunk` column not provided, trunk is unspecified"
        )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    phylogeny_df["ancestor_is_trunk"] = phylogeny_df.loc[
        phylogeny_df["ancestor_id"], "is_trunk"
    ].to_numpy()

    if np.any(phylogeny_df["is_trunk"] & ~phylogeny_df["ancestor_is_trunk"]):
        raise ValueError("specified trunk is non-contiguous")

    if phylogeny_df["is_trunk"].sum() == 0:
        return phylogeny_df

    trunk_df = phylogeny_df.loc[phylogeny_df["is_trunk"]]
    trunk_df = alifestd_mark_oldest_root(trunk_df, mutate=True)

    if "ancestor_id" in phylogeny_df:
        phylogeny_df.loc[
            phylogeny_df["ancestor_is_trunk"], "ancestor_id"
        ] = phylogeny_df.loc[phylogeny_df["ancestor_is_trunk"], "id"]

    if "ancestor_list" in phylogeny_df:
        phylogeny_df.loc[
            phylogeny_df["ancestor_is_trunk"], "ancestor_list"
        ] = "[none]"

    res = phylogeny_df.loc[~phylogeny_df["is_trunk"]].reset_index(drop=True)

    assert res["is_trunk"].sum() == 0
    return res



from ._policy_panel_animate import policy_panel_animate
from ._stratum_retention_animate import stratum_retention_animate

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "policy_panel_animate",
    "stratum_retention_animate",
]



import copy

import anytree
from keyname import keyname as kn
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import anytree_cardinality
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


@pytest.fixture
def trie():
    # (rank None, diff None) @ B6A
    # └── (rank 0, diff 2) @ BtQ
    #     ├── (rank 3, diff 31) @ jg
    #     |   └── space @ 0B
    #     └── (rank 3, diff 1) @ BUQ
    #         ├── spice @ 0A
    #         ├── sweet @ xA
    #         ├── (rank 4, diff 17) @ BxA
    #         │   └── (rank 5, diff 11) @ Gg
    #         │       ├── (rank 6, diff 19) @ Biw
    #         │       |  └── bank @ B-Q
    #         │       └── (rank 6, diff 21) @ HA
    #         │           └── baa @ B-Q
    #         ├── (rank 4, diff 16) @ B7Q
    #         │   └── (rank 5, diff 11) @ sA
    #         │       └── foo @ BtQ
    #         └── (rank 4, diff 17) @ Brw
    #             ├── (rank 6, diff 19) @ BLg
    #             │   ├── bump @ B0g
    #             │   ├── baz @ B0g
    #             │   └── barf @ 901
    #             └── (rank 6, diff 19) @ BoA
    #                 └── bar @ B8A

    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    genesis_node = impl.TrieInnerNode(rank=0, differentia=2, parent=root)
    sibling_node = impl.TrieInnerNode(
        rank=3, differentia=31, parent=genesis_node
    )
    impl.TrieLeafNode(sibling_node, "space")
    focal_node = impl.TrieInnerNode(rank=3, differentia=1, parent=genesis_node)
    impl.TrieLeafNode(focal_node, "spice")
    impl.TrieLeafNode(focal_node, "sweet")
    child1 = impl.TrieInnerNode(rank=4, differentia=17, parent=focal_node)
    child2 = impl.TrieInnerNode(rank=4, differentia=16, parent=focal_node)
    child3 = impl.TrieInnerNode(rank=4, differentia=17, parent=focal_node)
    grandchild1 = impl.TrieInnerNode(rank=5, differentia=11, parent=child1)
    grandchild2 = impl.TrieInnerNode(rank=5, differentia=11, parent=child2)
    impl.TrieLeafNode(grandchild2, "foo")
    greatgrandchild0 = impl.TrieInnerNode(
        rank=6, differentia=19, parent=grandchild1
    )
    impl.TrieLeafNode(greatgrandchild0, "bank")
    greatgrandchild1 = impl.TrieInnerNode(
        rank=6, differentia=21, parent=grandchild1
    )
    impl.TrieLeafNode(greatgrandchild1, "baa")
    greatgrandchild2 = impl.TrieInnerNode(
        rank=6, differentia=19, parent=child3
    )
    impl.TrieLeafNode(greatgrandchild2, "bump")
    impl.TrieLeafNode(greatgrandchild2, "baz")
    impl.TrieLeafNode(greatgrandchild2, "barf")
    greatgrandchild3 = impl.TrieInnerNode(
        rank=6, differentia=19, parent=child3
    )
    impl.TrieLeafNode(greatgrandchild3, "bar")

    return root


def test_base_class():
    assert issubclass(
        hstrat.PeelBackConjoinedLeavesTriePostprocessor,
        detail.TriePostprocessorBase,
    )


def test_singleton():
    root_node = impl.TrieInnerNode(rank=None, differentia=None)
    inner_node = impl.TrieInnerNode(rank=1, differentia=2, parent=root_node)
    impl.TrieLeafNode(taxon_label="bar", parent=inner_node)

    processed_root = hstrat.PeelBackConjoinedLeavesTriePostprocessor()(
        root_node,
        p_differentia_collision=0.25,
        mutate=False,
    )
    assert str(
        anytree.RenderTree(processed_root).by_attr("taxon_label")
    ) == str(anytree.RenderTree(root_node).by_attr("taxon_label"))


def test_twins():
    root_node = impl.TrieInnerNode(rank=None, differentia=None)
    root_node.test_name = "root"
    inner_node = impl.TrieInnerNode(rank=0, differentia=2, parent=root_node)
    inner_node.test_name = "inner"
    leaf_node1 = impl.TrieLeafNode(taxon_label="bar", parent=inner_node)
    leaf_node1.test_name = "leaf_bar"
    leaf_node2 = impl.TrieLeafNode(taxon_label="baz", parent=inner_node)
    leaf_node2.test_name = "leaf_baz"

    expected_root = impl.TrieInnerNode(rank=None, differentia=None)
    expected_root.test_name = "root"
    expected_inner1 = impl.TrieInnerNode(
        rank=0, differentia=2, parent=expected_root
    )
    expected_inner1.test_name = "inner"
    expected_inner2 = impl.TrieInnerNode(
        rank=0, differentia=2, parent=expected_root
    )
    expected_inner2.test_name = "inner"
    leaf_node1 = impl.TrieLeafNode(taxon_label="baz", parent=expected_inner1)
    leaf_node1.test_name = "leaf_baz"
    leaf_node2 = impl.TrieLeafNode(taxon_label="bar", parent=expected_inner2)
    leaf_node2.test_name = "leaf_bar"

    processed_root = hstrat.PeelBackConjoinedLeavesTriePostprocessor()(
        root_node,
        p_differentia_collision=0.25,
        mutate=False,
    )
    assert str(
        anytree.RenderTree(processed_root).by_attr("taxon_label")
    ) != str(anytree.RenderTree(root_node).by_attr("taxon_label"))
    assert str(anytree.RenderTree(processed_root).by_attr("test_name")) == str(
        anytree.RenderTree(expected_root).by_attr("test_name")
    )


def test_mutate(trie):
    root_node = impl.TrieInnerNode(rank=None, differentia=None)
    inner_node = impl.TrieInnerNode(rank=0, differentia=2, parent=root_node)
    impl.TrieLeafNode(taxon_label="bar", parent=inner_node)
    impl.TrieLeafNode(taxon_label="baz", parent=inner_node)

    root_bak = copy.deepcopy(root_node)

    hstrat.PeelBackConjoinedLeavesTriePostprocessor()(
        root_node,
        p_differentia_collision=0.25,
        mutate=False,
    )

    assert str(anytree.RenderTree(root_bak).by_attr("taxon_label")) == str(
        anytree.RenderTree(root_node).by_attr("taxon_label")
    )


def test_fixture(trie):
    cardinality_before = anytree_cardinality(trie)
    leaf_cardinality_before = len(trie.leaves)
    processed_trie = hstrat.PeelBackConjoinedLeavesTriePostprocessor()(
        trie,
        p_differentia_collision=0.25,
        mutate=False,
    )

    assert anytree_cardinality(processed_trie) == cardinality_before + 3

    assert len(trie.leaves) == leaf_cardinality_before
    for leaf in processed_trie.leaves:
        assert sum(1 for __ in leaf.parent.outer_children) == 1

    for leaf_taxon in "bump", "baz", "barf":
        leaf_node = next(
            node
            for node in anytree.PostOrderIter(processed_trie)
            if node.taxon_label == leaf_taxon
        )
        assert len(leaf_node.parent.children) == 1
        assert leaf_node.parent.IsAnOriginationOfAllele(rank=6, differentia=19)
        assert len(leaf_node.parent.parent.children) == 4

    for leaf_taxon in "spice", "sweet":
        leaf_node = next(
            node
            for node in anytree.PostOrderIter(processed_trie)
            if node.taxon_label == leaf_taxon
        )
        assert leaf_node.parent.IsAnOriginationOfAllele(rank=3, differentia=1)

    for node in anytree.PostOrderIter(processed_trie):
        attrs = kn.unpack(node.taxon_label)
        if "uid" in attrs:
            del attrs["uid"]
        node.test_name = kn.pack(attrs).rstrip("=")
        node.children = sorted(node.children, key=lambda x: x.test_name)

    assert str(anytree.RenderTree(processed_trie).by_attr("test_name")) == (
        """Root
└── Inner=+d=C+r=0
    ├── Inner=+d=B+r=3
    │   ├── Inner=+d=Q+r=4
    │   │   └── Inner=+d=L+r=5
    │   │       └── foo
    │   ├── Inner=+d=R+r=4
    │   │   └── Inner=+d=L+r=5
    │   │       ├── Inner=+d=T+r=6
    │   │       │   └── bank
    │   │       └── Inner=+d=V+r=6
    │   │           └── baa
    │   ├── Inner=+d=R+r=4
    │   │   ├── Inner=+d=T+r=6
    │   │   │   └── barf
    │   │   ├── Inner=+d=T+r=6
    │   │   │   └── bar
    │   │   ├── Inner=+d=T+r=6
    │   │   │   └── bump
    │   │   └── Inner=+d=T+r=6
    │   │       └── baz
    │   └── sweet
    ├── Inner=+d=B+r=3
    │   └── spice
    └── Inner=+d=f+r=3
        └── space"""
    )



#!/usr/bin/env python3

import random
import typing

from _SimpleGenomeAnnotatedWithDenseRetention import (
    SimpleGenomeAnnotatedWithDenseRetention,
)

# see here for example of how to incorporate hstrat into a custom genome
from _SimpleGenomeAnnotatedWithSparseRetention import (
    SimpleGenomeAnnotatedWithSparseRetention,
)

from hstrat import hstrat


def evolve_drift_synchronous(GenomeType: typing.Type) -> typing.List:
    population = [GenomeType() for __ in range(10)]
    # synchronous generations
    for generation in range(100):
        parents = random.choices(population, k=len(population))
        population = [parent.CreateOffspring() for parent in parents]

    return population


def evolve_tournament_synchronous(GenomeType: typing.Type) -> typing.List:
    population = [GenomeType() for __ in range(20)]

    # asynchronous generations
    for generation in range(100):
        # tournament selection with tournament size 7
        parents = [
            min(
                random.sample(population, k=7),
                key=lambda genome: genome.content,
            )
            for __ in population
        ]
        population = [parent.CreateOffspring() for parent in parents]

    return population


def evolve_tournament_asynchronous(GenomeType: typing.Type) -> typing.List:
    population = [GenomeType() for __ in range(20)]

    # asynchronous generations
    for generation in range(100 * 20):
        # tournament selection with tournament size 7
        random.shuffle(population)
        population[0] = min(
            random.sample(population, k=7),
            key=lambda genome: genome.content,
        ).CreateOffspring()

    return population


if __name__ == "__main__":
    for genome_type in (
        SimpleGenomeAnnotatedWithDenseRetention,
        SimpleGenomeAnnotatedWithSparseRetention,
    ):
        for experiment in (
            evolve_drift_synchronous,
            evolve_tournament_synchronous,
            evolve_tournament_asynchronous,
        ):
            print("genome type", genome_type.__name__)
            print("experiment", experiment.__name__)

            extant_population = experiment(genome_type)
            min_generations = min(
                genome.annotation.GetNumStrataDeposited()
                for genome in extant_population
            )
            print("   min generations elapsed", min_generations)
            max_generations = max(
                genome.annotation.GetNumStrataDeposited()
                for genome in extant_population
            )
            print("   max generations elapsed", max_generations)

            mrca_lb, mrca_ub = hstrat.calc_rank_of_mrca_bounds_among(
                (genome.annotation for genome in extant_population),
                prior="arbitrary",
            )
            print(
                "   population MRCA estmated between generation",
                mrca_lb,
                "and",
                mrca_ub,
            )

            print()



import typing

import anytree


def AnyTreeAscendingIter(
    node: anytree.AnyNode,
) -> typing.Iterator[anytree.AnyNode]:
    """Iterate over the ancestors of a given node in ascending order.

    Parameters
    ----------
    node : anytree.AnyNode
        A node in an anytree tree.

    Yields
    ------
    anytree.AnyNode
        The sequence of anytree.AnyNode objects representing the lineage of
        the given node in ascending order, starting from the node itself.

    Examples
    --------
    >>> root = anytree.Node("A")
    >>> b = anytree.Node("B", parent=root)
    >>> c = anytree.Node("C", parent=b)
    >>> d = anytree.Node("D", parent=b)
    >>> e = anytree.Node("E", parent=c)
    >>> f = anytree.Node("F", parent=d)
    >>> g = anytree.Node("G", parent=d)
    >>> list(AnyTreeAscendingIter(e))
    [Node('/A/B/C/E'), Node('/A/B/C'), Node('/A/B'), Node('/A')]
    """
    while node is not None:
        yield node
        node = node.parent



import functools
import random

import pytest
from tqdm import tqdm

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_pop_to_dataframe(retention_policy: object):
    pop = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]
    for __ in range(100):
        target = random.randrange(len(pop))
        pop[target] = random.choice(pop).CloneDescendant()

    df = hstrat.pop_to_dataframe(
        pop, progress_wrap=functools.partial(tqdm, disable=True)
    )

    assert len(df) == sum(column.GetNumStrataRetained() for column in pop)
    for col_idx, col_df in df.groupby("column"):
        assert (
            col_df.drop(
                "column",
                axis=1,
            )
            .reset_index(
                drop=True,
            )
            .equals(
                hstrat.col_to_dataframe(pop[col_idx]),
            )
        )



from collections import deque
import typing

from ...genome_instrumentation import HereditaryStratigraphicColumn
from .._calc_min_implausible_spurious_consecutive_differentia_collisions_between import (
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
)
from ._iter_ranks_of_retained_commonality_between import (
    iter_ranks_of_retained_commonality_between,
)


def calc_rank_of_last_retained_commonality_between_generic(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
    confidence_level: float,
) -> typing.Optional[int]:
    """Find rank of strata commonality before first strata disparity.

    Implementation detail with general-case implementation.
    """
    # we need to keep track of enough ranks of last-seen common strata so
    # that we can discount this many (minus 1) as potentially occuring due
    # to spurious differentia collisions
    assert (
        first.GetStratumDifferentiaBitWidth()
        == second.GetStratumDifferentiaBitWidth()
    )
    collision_implausibility_threshold = calc_min_implausible_spurious_consecutive_differentia_collisions_between(
        first, second, significance_level=1.0 - confidence_level
    )
    assert collision_implausibility_threshold > 0
    # holds up to n last-seen ranks with common strata,
    # with the newest last-seen rank at the front (index 0)
    # and the up to nth last-seen rank at the back (index -1)
    preceding_common_strata_ranks = deque(
        iter_ranks_of_retained_commonality_between(
            first,
            second,
            first_start_idx=first_start_idx,
            second_start_idx=second_start_idx,
        ),
        collision_implausibility_threshold,
    )
    if len(preceding_common_strata_ranks) < collision_implausibility_threshold:
        # not enough common strata to discount the possibility all
        # are spurious collisions with respect to the given confidence
        # level; conservatively conclude there is no common ancestor
        return None
    else:
        # discount collision_implausibility_threshold - 1 common strata
        # as potential spurious differentia collisions
        assert preceding_common_strata_ranks[0] >= 0
        return preceding_common_strata_ranks[0]



"""Implementation helpers."""

from ._DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH import (
    DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH,
)
from ._col_records_from_pop_records import col_records_from_pop_records
from ._policy_from_record import policy_from_record
from ._stringify_packed_differentia_bytes import (
    stringify_packed_differentia_bytes,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "col_records_from_pop_records",
    "DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH",
    "stringify_packed_differentia_bytes",
    "policy_from_record",
]



from ._surface_postprocess_trie import surface_postprocess_trie
from ._surface_unpack_reconstruct import surface_unpack_reconstruct

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "surface_unpack_reconstruct",
    "surface_postprocess_trie",
]



def bit_ceil(n: int) -> int:
    """Calculate the smallest power of 2 not smaller than a
    non-negative integer n."""
    assert n >= 0
    if n:
        # see https://stackoverflow.com/a/14267825/17332200
        # cast to int to make robust to numpy.int32, numpy.int64, etc.
        exp = int(n - 1).bit_length()
        res = 1 << exp
        return res
    else:
        return 1



"""Implementation helpers."""

from ._calc_rank_of_first_retained_disparity_between import (
    calc_rank_of_first_retained_disparity_between,
)
from ._calc_rank_of_last_retained_commonality_between import (
    calc_rank_of_last_retained_commonality_between,
)
from ._calc_rank_of_parity_segue_between import (
    calc_rank_of_parity_segue_between,
)
from ._calc_rank_of_parity_segue_between_naive import (
    calc_rank_of_parity_segue_between_naive,
)
from ._iter_mutual_rank_indices import iter_mutual_rank_indices
from ._iter_mutual_ranks import iter_mutual_ranks
from ._iter_ranks_of_retained_commonality_between import (
    iter_ranks_of_retained_commonality_between,
)

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "calc_rank_of_first_retained_disparity_between",
    "calc_rank_of_last_retained_commonality_between",
    "calc_rank_of_parity_segue_between",
    "iter_mutual_rank_indices",
    "iter_mutual_ranks",
    "calc_rank_of_parity_segue_between_naive",
    "iter_ranks_of_retained_commonality_between",
]



import typing


def raises(callable_: typing.Callable, exception: BaseException) -> bool:
    """Test if a `callable_` raises a `exception` when called.

    Parameters
    ----------
    callable_ : typing.Callable
        The callable to be tested for raising the specified exception.
    exception : BaseException
        The exception that is expected to be raised by the callable.

    Returns
    -------
    bool
        True if the callable raises the specified exception; False otherwise.

    Examples
    --------
    >>> def my_func():
    ...     raise ValueError("test")
    >>> raises(my_func, ValueError)
    True

    >>> def my_func():
    ...     return True
    >>> raises(my_func, ValueError)
    False
    """
    try:
        callable_()
        return False
    except exception:
        return True



from ._calc_geom_seq_nth_root_transition_rank import (
    calc_geom_seq_nth_root_transition_rank,
)
from ._calc_provided_degree import calc_provided_degree
from ._calc_provided_resolution import calc_provided_resolution
from ._pick_policy import pick_policy

__all__ = [
    "calc_geom_seq_nth_root_transition_rank",
    "calc_provided_degree",
    "calc_provided_resolution",
    "pick_policy",
]



import anytree


def anytree_has_grandparent(node: anytree.node) -> bool:
    """Check whether the given node is located at or below level two of its
    tree.

    Parameters
    ----------
    node : anytree.node
        The node to check.

    Returns
    -------
    bool
        True if the node has a grandparent in the tree, False otherwise.
    """
    return node.parent is not None and node.parent.parent is not None



import numbers
import typing

from ._unzip import unzip


def intersect_ranges(
    ranges: typing.Iterable[typing.Tuple[numbers.Real, numbers.Real]]
) -> typing.Optional[typing.Tuple[numbers.Real, numbers.Real]]:
    """Calculate the intersection of a collection of ranges.

    Parameters
    ----------
    ranges : iterable of tuples of real numbers
        A list of ranges represented as tuples of two real numbers,
        where the first number is the beginning of the range and the
        second number is the end of the range.

    Returns
    -------
    tuple of real numbers or None
        A tuple representing the intersection of the input ranges,
        where the first number is the beginning of the intersection and
        the second number is the end of the intersection. If there is no
        intersection, returns None.

    Examples
    --------
    >>> intersect_ranges([(0, 5), (3, 9), (3, 10)])
    (3, 5)

    >>> intersect_ranges([(1, 5)])
    (1, 5)

    >>> intersect_ranges([])
    None

    >>> intersect_ranges([(1, 3), (4, 7), (8, 10)])
    None
    """

    try:
        begins, ends = unzip(ranges)
    except ValueError:
        return None

    begin = max(begins)
    end = min(ends)

    return (begin, end) if begin < end else None



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcNumStrataRetainedUpperBound:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcNumStrataRetainedUpperBound",
        policy_spec: typing.Optional[PolicySpec] = None,
    ) -> None:
        pass

    def __eq__(
        self: "CalcNumStrataRetainedUpperBound",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcNumStrataRetainedUpperBound",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> int:
        """At most, how many strata are retained after n deposited? Inclusive."""
        spec = policy.GetSpec()

        return min(
            spec.GetDepthProportionalResolution() * 2 + 1,
            num_strata_deposited,
        )



import typing

from .._detail import PolicySpecBase


class PolicySpec(PolicySpecBase):
    """Contains all policy parameters, if any."""

    _depth_proportional_resolution: int

    def __init__(
        self: "PolicySpec",
        depth_proportional_resolution: int,
    ):
        """Construct the policy spec.

        Parameters
        ----------
        depth_proportional_resolution : int
            The desired minimum number of intervals for the rank of the MRCA to
            be able to be distinguished between. The uncertainty of MRCA
            rank estimates provided under the tapered depth-proportional
            resolution policy will scale as total number of strata deposited
            divided by depth_proportional_resolution.
        """
        assert depth_proportional_resolution > 0
        self._depth_proportional_resolution = depth_proportional_resolution

    def __eq__(self: "PolicySpec", other: typing.Any) -> bool:
        return isinstance(other, self.__class__) and (
            self._depth_proportional_resolution,
        ) == (other._depth_proportional_resolution,)

    def __repr__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoIdentifier()
        }.{
            PolicySpec.__qualname__
        }(depth_proportional_resolution={
            self._depth_proportional_resolution
        })"""

    def __str__(self: "PolicySpec") -> str:
        return f"""{
            self.GetAlgoTitle()
        } (resolution: {
            self._depth_proportional_resolution
        })"""

    def GetEvalCtor(self: "PolicySpec") -> str:
        return f"hstrat.{self!r}"

    def GetDepthProportionalResolution(self: "PolicySpec") -> int:
        return self._depth_proportional_resolution

    @staticmethod
    def GetAlgoIdentifier() -> str:
        """Get programatic name for underlying retention algorithm."""
        return __package__.split(".")[-1]

    @staticmethod
    def GetAlgoTitle() -> str:
        """Get human-readable name for underlying retention algorithm."""
        return (
            "Tapered Depth-proportional Resolution Stratum Retention Algorithm"
        )



"""Retain all strata to ensure perfect resolution for MRCA estimation.

The perfect resolution policy retains all strata. So, comparisons between
two columns under this policy will detect MRCA rank with zero
uncertainty. So, MRCA rank estimate uncertainty scales as O(1) with respect
to the greater number of strata deposited on either column.

Under the perfect resolution policy, the number of strata retained (i.e.,
space complexity) scales as O(n) with respect to the number of strata
deposited.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import typing
import warnings

import mpmath as mp


def calc_common_ratio(
    degree: int,
    num_strata_deposited: int,
) -> typing.Union[float, mp.mpf]:
    """What should the base of the exponential distribution of retained
    ranks be?"""
    # this try-except isn't necessary for the test cases
    # and the bounds of normal usage, but provides better resiliency
    # for extreme magnitude num_strata_deposited which may arise from time
    # to time due to trickle down to this function from within a doubling
    # search
    # the mpf type (artibtrary-precision floating-point) will propagate
    # forward through subsequent computations
    try:
        num_strata_deposited = float(num_strata_deposited)
    except OverflowError:
        warnings.warn(
            "OverflowError converting num_strata_deposited to float, "
            "converting to mpmath.mpf instead.",
        )
        num_strata_deposited = mp.mpf(num_strata_deposited)

    # base ** degree == num_strata_deposited
    # take the degree'th root of each side...
    return num_strata_deposited ** (1 / degree)



#!/usr/bin/env python3

from hstrat import hstrat

if __name__ == "__main__":

    for policy in (
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(3),
    ):
        print()
        print(policy)

        col = hstrat.HereditaryStratigraphicColumn(
            stratum_retention_policy=policy,
        )
        for __ in range(20):
            col.DepositStratum()

        print(hstrat.col_to_ascii(col))



import os
import pathlib
import subprocess

import pytest

assets = os.path.join(os.path.dirname(os.path.abspath(__file__)), "assets")


def test_alifestd_as_newick_asexual_cli_version():
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_as_newick_asexual",
            "--version",
        ],
        check=True,
    )


@pytest.mark.parametrize(
    "input_file",
    [
        "nk_ecoeaselection.csv",
        "nk_lexicaseselection.csv",
        "nk_tournamentselection.csv",
    ],
)
def test_alifestd_as_newick_asexual_cli_csv(input_file: str):
    output_file = f"/tmp/hstrat-{input_file}.newick"
    pathlib.Path(output_file).unlink(missing_ok=True)
    subprocess.run(
        [
            "python3",
            "-m",
            "hstrat._auxiliary_lib._alifestd_as_newick_asexual",
            "--input-file",
            f"{assets}/{input_file}",
            "-o",
            output_file,
        ],
        check=True,
    )
    assert os.path.exists(output_file)



#!/usr/bin/env python3

import os
import random

import opytional as opyt

from hstrat import hstrat

try:
    from Bio import Phylo as BioPhylo
except (ImportError, ModuleNotFoundError) as e:
    print("biopython required for tree reconstruction example")
    print("python3 -m pip install biopython")
    raise e

assets_path = os.path.join(os.path.dirname(__file__), "assets")

if __name__ == "__main__":
    print("loaded template phylogeny from assets/example.newick:")
    template_tree = BioPhylo.read(f"{assets_path}/example.newick", "newick")
    template_tree.rooted = True
    BioPhylo.draw_ascii(template_tree)
    for clade in template_tree.find_clades():
        clade.branch_length = int(opyt.or_value(clade.branch_length, 0))

    print("simulated hstrat column inheritance along template phylogeny:")
    extant_population = hstrat.descend_template_phylogeny_biopython(
        template_tree,
        seed_column=hstrat.HereditaryStratigraphicColumn(
            hstrat.recency_proportional_resolution_algo.Policy(5)
        ),
    )

    depth_lookup = template_tree.depths()
    clade_lookup = {v: k for k, v in depth_lookup.items()}
    for column in extant_population:
        generation = column.GetNumStrataDeposited() - 1
        clade = clade_lookup[generation]
        print(clade.name, f"{generation} generations ==>", column)

    print()
    print("a few pairwise comparisons via hstrat columns")
    for _rep in range(8):
        c1, c2 = random.sample(extant_population, 2)
        print(
            "hstrat.calc_rank_of_mrca_bounds_between "
            f"{clade_lookup[c1.GetNumStrataDeposited() - 1].name} & {clade_lookup[c2.GetNumStrataDeposited() - 1].name}: "
            f"{hstrat.calc_rank_of_mrca_bounds_between(c1, c2, prior='arbitrary')}"
        )



import typing
import warnings

import pandas as pd

from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_make_ancestor_id_col import alifestd_make_ancestor_id_col
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._all_unique import all_unique
from ._is_subset import is_subset


def _validate_ancestors_asexual(
    phylogeny_df: pd.DataFrame, mutate: bool, warn: typing.Callable
) -> bool:
    if "ancestor_id" not in phylogeny_df:
        if not mutate:
            phylogeny_df = phylogeny_df.copy()
        phylogeny_df["ancestor_id"] = alifestd_make_ancestor_id_col(
            phylogeny_df["id"], phylogeny_df["ancestor_list"]
        )
    else:
        ok_ancestor_list_mask = (
            phylogeny_df["ancestor_list"]
            .astype("str")
            .str.lower()
            .replace("[]", "[none]")
            .to_numpy()
            == alifestd_make_ancestor_list_col(
                phylogeny_df["id"], phylogeny_df["ancestor_id"]
            ).to_numpy()
        )
        if not (ok_ancestor_list_mask).all():
            example_row = phylogeny_df[~ok_ancestor_list_mask].iloc[0]
            warn(
                "alifestd_validate asexual: "
                "ancestor_list nonequivalent to "
                "alifestd_make_ancestor_list_col result"
                f"with {(~ok_ancestor_list_mask).sum()} violations, "
                f"for example\n{example_row}"
            )
            return False

    ancestor_ids_are_subset = is_subset(
        phylogeny_df["ancestor_id"].to_numpy(),
        phylogeny_df["id"].astype("int").to_numpy(),
    )
    if not ancestor_ids_are_subset:
        warn(
            "alifestd_validate asexual: "
            "ancestor_id values are not a subset of id values",
        )
    return ancestor_ids_are_subset


def _validate_ancestors_sexual(
    phylogeny_df: pd.DataFrame,
    warn: typing.Callable,
) -> bool:
    ids = set(phylogeny_df["id"])
    ancestor_ids_are_subset = all(
        ancestor_id in ids
        for ancestor_list_str in phylogeny_df["ancestor_list"]
        for ancestor_id in alifestd_parse_ancestor_ids(ancestor_list_str)
        if ancestor_id is not None
    )
    if not ancestor_ids_are_subset:
        warn(
            "alifestd_validate sexual: "
            "ancestor_id values are not a subset of id values",
        )
    return ancestor_ids_are_subset


def _alifestd_validate(
    phylogeny_df: pd.DataFrame,
    mutate: bool,
    warn: typing.Callable,
) -> bool:
    has_mandatory_columns = (
        "id" in phylogeny_df and "ancestor_list" in phylogeny_df
    )
    if not has_mandatory_columns:
        warn("alifestd_validate: missing mandatory columns")
        return False

    ids_valid = (
        len(phylogeny_df) == 0
        or pd.api.types.is_integer_dtype(phylogeny_df["id"])
        and (phylogeny_df["id"] >= 0).all()
        and all_unique(phylogeny_df["id"].to_numpy())
    )
    if not ids_valid:
        warn(
            "alifestd_validate: invalid id detected, "
            f"{phylogeny_df['id'].dtype=}, "
            f"{(phylogeny_df['id'] >= 0).all()=}, "
            f"{all_unique(phylogeny_df['id'].to_numpy())=}",
        )
        return False

    ancestor_lists_syntax_valid = (
        len(phylogeny_df) == 0
        or pd.api.types.is_string_dtype(phylogeny_df["ancestor_list"])
        and phylogeny_df["ancestor_list"].str.startswith("[").all()
        and phylogeny_df["ancestor_list"].str.endswith("]").all()
    )
    if not ancestor_lists_syntax_valid:
        warn("alifestd_validate: invalid ancestor_list syntax detected")
        return False

    return (
        _validate_ancestors_asexual(phylogeny_df, mutate, warn)
        if alifestd_is_asexual(phylogeny_df)
        else _validate_ancestors_sexual(phylogeny_df, warn)
    )


def alifestd_validate(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    diagnose: bool = True,
) -> bool:
    """Is the phylogeny compliant to alife data standards?

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If `diagnose` is set, the failing validation subcheck will warn.
    """

    warn = warnings.warn if diagnose else lambda x: x

    try:
        return _alifestd_validate(phylogeny_df, mutate, warn)
    except Exception as exception:
        warn(f"alifestd_validate: {exception=} occured")
        return False



# adapted from https://cs.stackexchange.com/a/65744
def parse_from_numeral_system(digits: str, alphabet: str) -> int:
    """Compute the integer value represented by string 'digits' in a custom
    base.

    Parameters
    ----------
    digits : str
        A string of digits in the custom base.
    alphabet : str
        A string of unique characters representing the digits in the custom
        base.

        The i-th character in the alphabet corresponds to the i-th digit
        in the base.

    Returns
    -------
    int
        The integer value of the number represented by the given digits in the
        custom base.

    Raises
    ------
    AssertionError
        If the alphabet contains repeated characters.

    Examples
    --------
    >>> parse_from_numeral_system('a3', '0123456789abcdef')
    163
    >>> parse_from_numeral_system('52', '01234567')
    42
    >>> parse_from_numeral_system('111', '01')
    7
    """
    assert len(alphabet) == len(set(alphabet))
    b = len(alphabet)
    n = 0
    for d in digits:
        n = b * n + alphabet.index(d)
    return n



import typing

from ..genome_instrumentation import HereditaryStratigraphicColumn
from ._col_to_packet import col_to_packet
from ._impl import DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH


def col_to_int(
    column: HereditaryStratigraphicColumn,
    num_strata_deposited_byte_order: typing.Literal["big", "little"] = "big",
    num_strata_deposited_byte_width: int = (
        DEFAULT_PACKET_NUM_STRATA_DEPOSITED_BYTE_WIDTH
    ),
) -> int:
    """Serialize a `HereditaryStratigraphicColumn` to a binary representation
    as a Python int.

    Integer representation is packet binary representation plus a sentry bit at
    the most significant bit position, to prevent loss of leading zero bits.

    Uses big endian byte order.
    """

    if (
        not column._CanOmitStratumDepositionRank()
        or column.HasAnyAnnotations()
    ):
        raise NotImplementedError()

    buffer = col_to_packet(
        column=column,
        num_strata_deposited_byte_order=num_strata_deposited_byte_order,
        num_strata_deposited_byte_width=num_strata_deposited_byte_width,
    )
    buffer_num_bits = len(buffer) * 8
    sentry_bit = 1 << buffer_num_bits
    assert sentry_bit.bit_length() % 8 == 1
    semantic_value = int.from_bytes(buffer, byteorder="big", signed=False)
    assert semantic_value.bit_length() <= buffer_num_bits
    assert semantic_value & sentry_bit == 0

    return semantic_value | sentry_bit



import pytest

from hstrat import hstrat


@pytest.mark.filterwarnings(
    "ignore:Animation was deleted without rendering anything."
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.geom_seq_nth_root_algo.Policy(4),
    ],
)
def test_one(policy):
    hstrat.policy_panel_animate(policy, 10,).to_html5_video(
        embed_limit=0
    )  # silence mpl unused animation warning


@pytest.mark.heavy
def test_doc_animations(docplot_policy):
    hstrat.policy_panel_animate(
        docplot_policy,
        256,
        save_as="gif",
    )



import typing

import pandas as pd

_supported_iterables = tuple, set, list, frozenset
_supported_mappings = dict


def coerce_to_pandas(obj: typing.Any, *, recurse: bool = False) -> typing.Any:
    """
    If a Polars type is detected, coerce it to corresponding Pandas type.
    """
    if hasattr(obj, "__dataframe__"):
        return pd.api.interchange.from_dataframe(obj, allow_copy=True)
    elif hasattr(obj, "to_pandas"):
        return obj.to_pandas()  # pyarrow is required for this operation
    elif recurse and isinstance(obj, _supported_iterables):
        return type(obj)(
            map(lambda x: coerce_to_pandas(x, recurse=recurse), obj)
        )
    elif recurse and isinstance(
        obj, _supported_mappings
    ):  # includes defaultdict etc
        return type(obj)(
            {k: coerce_to_pandas(v, recurse=recurse) for k, v in obj.items()},
        )
    else:
        return obj



import functools
import itertools as it
import typing

import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_estimate_triplet_distance_asexual,
    alifestd_make_ancestor_list_col,
    alifestd_make_empty,
    alifestd_to_working_format,
    alifestd_validate,
)


def test_empty():
    assert 0.0 == alifestd_estimate_triplet_distance_asexual(
        alifestd_make_empty(), alifestd_make_empty(), "id"
    )


def test_undersize():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2],
            "taxon_label": [4, 5, 6],
            "ancestor_id": [0, 0, 0],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df[:1], df[:2], df:
        assert (
            alifestd_estimate_triplet_distance_asexual(
                xdf, xdf.sample(frac=1), "taxon_label", detail=False
            )
            == 0.0
        )


@pytest.mark.parametrize(
    "df",
    [
        pd.DataFrame(
            {
                "id": [1, 2, 3],
                "taxon_label": ["a", "b", "c"],
                "ancestor_id": [1, 1, 1],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3],
                "taxon_label": [0, 1, 2, 3],
                "ancestor_id": [0, 0, 0, 0],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5],
                "taxon_label": ["0", "1", "2", "3", "4", "5"],
                "ancestor_id": [0, 0, 0, 0, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [9, 1, 2, 3, 4, 5],
                "taxon_label": ["9", "1", "2", "3", "4", "5"],
                "ancestor_id": [9, 9, 9, 9, 1, 2],
            },
        ),
        pd.DataFrame(
            {
                "id": [0, 1, 2, 3, 4, 5, 6],
                "taxon_label": ["0", "1", "2", "3", "4", "5", "6"],
                "ancestor_id": [0, 0, 0, 0, 1, 2, 3],
            },
        ),
    ],
)
def test_polytomy_identical(df: pd.DataFrame):
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for adf, bdf in (df, df), (df, df.sample(frac=1)):
        est, ci, n = alifestd_estimate_triplet_distance_asexual(
            adf, bdf, "taxon_label", confidence=0.95, detail=True
        )
        assert est == 0
        assert np.isclose(ci[0], 0)
        assert ci[1] >= ci[0]


@pytest.mark.parametrize(
    "strict",
    [
        True,
        False,
        *it.product([True, False], repeat=2),
    ],
)
def test_differing_wrong1(
    strict: typing.Union[bool, typing.Tuple[bool, bool]]
):
    adf = pd.DataFrame(
        {
            "id": reversed([9, 1, 2, 3, 4, 5]),
            "taxon_label": reversed(["0", "1", "2", "3", "4", "5"]),
            "ancestor_id": reversed([9, 9, 1, 2, 2, 1]),
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 5, 4],
            "taxon_label": ["0", "1", "2", "3", "5", "4"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf.sample(frac=1), "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est


@pytest.mark.parametrize(
    "strict", [True, False, *it.product([True, False], repeat=2)]
)
def test_differing_wrong2(
    strict: typing.Union[bool, typing.Tuple[bool, bool]]
):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 1, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf.sample(frac=1), "taxon_label", confidence=0.95, strict=strict
    )
    assert 1 == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf.sample(frac=1),
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        strict=strict,
    )
    assert 1 == est


@pytest.mark.parametrize("strict", [True, False])
def test_differing_polytomy(strict: bool):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "id", confidence=0.95, precision=0.05, strict=strict
    )
    assert bool(strict) == est

    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=strict,
    )
    assert bool(strict) == est


@pytest.mark.parametrize(
    "strict",
    [
        True,
        False,
        *it.product([True, False], repeat=2),
    ],
)
def test_differing_polytomy2(
    strict: typing.Union[bool, typing.Tuple[bool, bool]]
):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "taxon_label": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 3, 3, 2],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "taxon_label": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 3, 2, 3],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "id", confidence=0.85, precision=0.05, strict=strict
    )
    assert est

    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )
    assert est


@pytest.mark.parametrize(
    "strict",
    [
        True,
        False,
        *it.product([True, False], repeat=2),
    ],
)
def test_identical_polytomy1(
    strict: typing.Union[bool, typing.Tuple[bool, bool]]
):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "taxon_label": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 3, 3, 2],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8],
            "taxon_label": ["0", "1", "2", "3", "4", "5", "6", "7", "8"],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 3, 3, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "id", confidence=0.95, precision=0.05, strict=strict
    )
    assert not est

    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )
    assert not est


@pytest.mark.parametrize(
    "strict",
    [
        True,
        False,
        *it.product([True, False], repeat=2),
    ],
)
def test_differing_wrong_big(
    strict: typing.Union[bool, typing.Tuple[bool, bool]]
):
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "taxon_label": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 5, 5, 4, 4],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "taxon_label": [0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
            "ancestor_id": [0, 0, 1, 5, 2, 1, 2, 5, 4, 4],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf,
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=strict,
    )
    assert 0 < est < 1

    est = alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=strict,
    )
    assert 0 < est < 1


def test_differing_polytomy_asymmetrical_strict():
    adf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    adf["ancestor_list"] = alifestd_make_ancestor_list_col(
        adf["id"], adf["ancestor_id"]
    )
    bdf = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 2, 2],
        },
    )
    bdf["ancestor_list"] = alifestd_make_ancestor_list_col(
        bdf["id"], bdf["ancestor_id"]
    )
    assert not alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "id", confidence=0.95, precision=0.05, strict=(True, False)
    )
    assert alifestd_estimate_triplet_distance_asexual(
        adf, bdf, "id", confidence=0.95, precision=0.05, strict=(False, True)
    )

    assert alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=(True, True),
    )

    assert not alifestd_estimate_triplet_distance_asexual(
        adf,
        bdf.sample(frac=1),
        "taxon_label",
        confidence=0.95,
        precision=0.05,
        strict=(False, False),
    )


@functools.lru_cache
def make_ref_df(working_format: bool) -> pd.DataFrame:
    #        |
    #      __|_
    #    __|_ |
    #  __|_ | |
    # _|_ | | |
    # | | | | |
    # A B C D E
    ref_df = pd.DataFrame.from_records(
        [
            {"id": 0, "taxon_label": "A", "ancestor_id": 5},
            {"id": 1, "taxon_label": "B", "ancestor_id": 5},
            {"id": 2, "taxon_label": "C", "ancestor_id": 6},
            {"id": 3, "taxon_label": "D", "ancestor_id": 7},
            {"id": 4, "taxon_label": "E", "ancestor_id": 8},
            {"id": 5, "taxon_label": "AB", "ancestor_id": 6},
            {"id": 6, "taxon_label": "ABC", "ancestor_id": 7},
            {"id": 7, "taxon_label": "ABCD", "ancestor_id": 8},
            {"id": 8, "taxon_label": "ABCDE", "ancestor_id": 8},
        ]
    )
    ref_df["ancestor_list"] = alifestd_make_ancestor_list_col(
        ref_df["id"], ref_df["ancestor_id"]
    )
    if working_format:
        ref_df = alifestd_to_working_format(ref_df)

    assert alifestd_validate(ref_df)
    assert not alifestd_estimate_triplet_distance_asexual(
        ref_df,
        ref_df.sample(frac=1),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=True,
    )
    return ref_df


@functools.lru_cache
def make_swap_df(working_format: bool) -> pd.DataFrame:
    #       |
    #    ___|__
    #  __|__  |
    # _|_ _|_ |
    # | | | | |
    # * B C D E
    # A
    swap_df = pd.DataFrame.from_records(
        [
            {"id": 0, "taxon_label": "A", "ancestor_id": 9},
            {"id": 1, "taxon_label": "B", "ancestor_id": 5},
            {"id": 2, "taxon_label": "C", "ancestor_id": 6},
            {"id": 3, "taxon_label": "D", "ancestor_id": 6},
            {"id": 4, "taxon_label": "E", "ancestor_id": 8},
            {"id": 5, "taxon_label": "AB", "ancestor_id": 7},
            {"id": 6, "taxon_label": "CD", "ancestor_id": 7},
            {"id": 7, "taxon_label": "ABCD", "ancestor_id": 8},
            {"id": 8, "taxon_label": "ABCDE", "ancestor_id": 8},
            {"id": 9, "taxon_label": "A*", "ancestor_id": 5},
        ]
    )
    swap_df["ancestor_list"] = alifestd_make_ancestor_list_col(
        swap_df["id"], swap_df["ancestor_id"]
    )
    if working_format:
        swap_df = alifestd_to_working_format(swap_df)

    assert alifestd_validate(swap_df)
    assert not alifestd_estimate_triplet_distance_asexual(
        swap_df,
        swap_df.sample(frac=1),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=True,
    )
    return swap_df


@functools.lru_cache
def make_polytomy_df(working_format: bool) -> pd.DataFrame:
    #       |
    #     __|__
    #  ___|__ |
    # _|_ | | |
    # | | | | |
    # A B C D E
    polytomy_df = pd.DataFrame.from_records(
        [
            {"id": 0, "taxon_label": "A", "ancestor_id": 5},
            {"id": 1, "taxon_label": "B", "ancestor_id": 5},
            {"id": 2, "taxon_label": "C", "ancestor_id": 7},
            {"id": 3, "taxon_label": "D", "ancestor_id": 7},
            {"id": 4, "taxon_label": "E", "ancestor_id": 8},
            {"id": 5, "taxon_label": "AB", "ancestor_id": 7},
            {"id": 7, "taxon_label": "ABCD", "ancestor_id": 8},
            {"id": 8, "taxon_label": "ABCDE", "ancestor_id": 8},
        ]
    )
    polytomy_df["ancestor_list"] = alifestd_make_ancestor_list_col(
        polytomy_df["id"], polytomy_df["ancestor_id"]
    )
    if working_format:
        polytomy_df = alifestd_to_working_format(polytomy_df)

    assert alifestd_validate(polytomy_df)
    assert not alifestd_estimate_triplet_distance_asexual(
        polytomy_df,
        polytomy_df.sample(frac=1),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=True,
    )
    return polytomy_df


@pytest.mark.parametrize(
    "strict",
    [
        True,
        False,
        *it.product([True, False], repeat=2),
    ],
)
@pytest.mark.parametrize(
    "working_format", [True, pytest.param(False, marks=pytest.mark.heavy)]
)
def test_sensitivity1(
    strict: typing.Union[bool, typing.Tuple[bool, bool]],
    working_format: bool,
):

    assert alifestd_estimate_triplet_distance_asexual(
        make_ref_df(working_format),
        make_swap_df(working_format),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )


@pytest.mark.parametrize(
    "strict",
    [
        True,
        (False, True),
    ],
)
@pytest.mark.parametrize(
    "working_format", [True, pytest.param(False, marks=pytest.mark.heavy)]
)
def test_sensitivity2(
    strict: typing.Union[bool, typing.Tuple[bool, bool]],
    working_format: bool,
):
    assert alifestd_estimate_triplet_distance_asexual(
        make_ref_df(working_format),
        make_polytomy_df(working_format),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )
    assert alifestd_estimate_triplet_distance_asexual(
        make_ref_df(working_format).sample(frac=1),
        make_polytomy_df(working_format).sample(frac=1),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )


@pytest.mark.parametrize(
    "strict",
    [
        False,
        (True, False),
    ],
)
@pytest.mark.parametrize(
    "working_format", [True, pytest.param(False, marks=pytest.mark.heavy)]
)
def test_sensitivity3(
    strict: typing.Union[bool, typing.Tuple[bool, bool]],
    working_format: bool,
):
    assert not alifestd_estimate_triplet_distance_asexual(
        make_ref_df(working_format),
        make_polytomy_df(working_format),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )


@pytest.mark.parametrize(
    "strict",
    [
        True,
        (False, True),
    ],
)
@pytest.mark.parametrize(
    "working_format", [True, pytest.param(False, marks=pytest.mark.heavy)]
)
def test_sensitivity4(
    strict: typing.Union[bool, typing.Tuple[bool, bool]],
    working_format: bool,
):
    assert alifestd_estimate_triplet_distance_asexual(
        make_swap_df(working_format),
        make_polytomy_df(working_format),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )


@pytest.mark.parametrize(
    "strict",
    [
        False,
        (False, True),
    ],
)
@pytest.mark.parametrize(
    "working_format", [True, pytest.param(False, marks=pytest.mark.heavy)]
)
def test_sensitivity5(
    strict: typing.Union[bool, typing.Tuple[bool, bool]],
    working_format: bool,
):
    assert not alifestd_estimate_triplet_distance_asexual(
        make_polytomy_df(working_format),
        make_swap_df(working_format),
        "taxon_label",
        confidence=0.85,
        precision=0.05,
        strict=strict,
    )



import argparse
import functools
import logging

from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli

from .._auxiliary_lib import (
    configure_prod_logging,
    format_cli_description,
    get_hstrat_version,
    log_context_duration,
)
from ._surface_unpack_reconstruct import surface_unpack_reconstruct

raw_message = """First component of raw interface to tree reconstruction for surface-based genome annotations.
Reads raw genome data from tabular data file(s) and writes reconstructed phylogeny data to output file in alife standard format.


End-users are recommended to prefer `hstrat.dataframe.build_tree`.
Performs no post-processing operations, including taxon `origin_time` estimation.
Surface-only implementation --- not compatible with HereditaryStratigraphicColumn genome annotations.


Input Schema: Required Columns
==============================
'data_hex' : string
    Raw genome data, with serialized dstream buffer and counter.
    Represented as a hexadecimal string.

'dstream_algo' : string or categorical
    Name of downstream curation algorithm used (e.g., 'dstream.steady_algo').

'dstream_storage_bitoffset' : integer
    Position of dstream buffer field in 'data_hex'.

'dstream_storage_bitwidth' : integer
    Size of dstream buffer field in 'data_hex'.

'dstream_T_bitoffset' : integer
    Position of dstream counter field ("rank”) in 'data_hex'.

'dstream_T_bitwidth' : integer
    Size of dstream counter field in 'data_hex'.

'dstream_S' : integer
    Capacity of dstream buffer, in number of data items (i.e., num differentia stored per annotation).


If required columns are not included in raw data, they must be created through CLI flags.
For example,
    --with-column 'pl.lit("dstream.tilted_algo").alias("dstream_algo").cast(pl.Categorical)'”
    --with-column 'pl.lit("genome_string").alias("data_hex")'


Input Schema: Optional Columns
==============================
'downstream_version' : string or categorical
    Version of downstream library used to curate annotation data.

'dstream_data_id' : integer
    Unique identifier for each data item. Will be set automatically if not provided.


Additional user-provided columns will be forwarded to phylogeny output.
For these columns, output rows for tip nodes are assigned values from corresponding genome row in original data.
Internal tree nodes will take null values in user-provided columns.


Output Schema
=============
'id' : integer
    Unique identifier for each taxon (RE alife standard format).

'ancestor_id' : integer
    Unique identifier for ancestor taxon (RE alife standard format).

'hstrat_rank' : floating point or integer
    Num generations elapsed for ancestral differentia (a.k.a. rank).
    Corresponds to `dstream_Tbar` for inner nodes.
    Corresponds `dstream_T` - 1 for leaf nodes.

'differentia_bitwidth' : integer
    Size of annotation differentiae, in bits.
    Corresponds to `dstream_value_bitwidth`.

'dstream_data_id' : integer
    Unique identifier for each genome in source dataframe.
    Set to source dataframe row index if not provided by end user.

'dstream_S' : pl.UInt32
    Capacity of dstream buffer used for hstrat surface, in number of data items (i.e., differentia values).


See Also
========
-m hstrat.dataframe.surface_postprocess_trie :
    Post-processes raw phylogeny data, including collapsing superfluous internal nodes and estimating taxon origin times.

-m hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col :
    Adds alife-standard `ancestor_list` column to phylogeny data.


Additional Notes
================
Behind the scenes, the following processing steps occur:
1. Data is loaded from user-provided tabular data file(s), with format inferred from filename extension (if not explicitly provided).

2. Any data transformations from CLI flags and arguments are applied.

3. Annotation data (i.e., downstream buffer and counter components of hstrat surface) are extracted from raw genome fields.
For details, see `python3 -m downstream.dataframe.unpack_data_packed --help`.

4. Differentia are unpacked from downstream buffer and data item origin times are decoded from generation counter and buffer size.
For details, see `python3 -m downstream.dataframe.explode_lookup_unpacked --help`.

5. A trie-based algorithm is used to reconstruct an estimate of phylogenetic history from annotation data.

6. Reconstructed phylogeny data is joined with user-provided columns from original data files.

7. Phylogeny data is saved in alife standard format to user-specified output file, with file type inferred from filename extension (if not explicitly provided).


To streamline memory and disk usage, consider using CLI flags to cast string columns with repeated data values to categorical, shrink data types, or drop superfluous columns.
The `--exploded-slice-size` flag may also be used to control memory usage during trie reconstruction.
Dataframe operations are conducted using polars and downstream operations may employ numba, both of which are capable of thread-based parallelism.
Environment variables POLARS_MAX_THREADS and NUMBA_NUM_THREADS may be used to tune thread usage.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=format_cli_description(raw_message),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _add_parser_base(
        parser=parser,
        dfcli_module="hstrat.dataframe.surface_unpack_reconstruct",
        dfcli_version=get_hstrat_version(),
    )
    parser.add_argument(
        "--exploded-slice-size",
        type=int,
        default=1_000_000,
        help="Number of rows to process at once. Low values reduce memory use.",
    )
    return parser


if __name__ == "__main__":
    configure_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    with log_context_duration(
        "hstrat.dataframe.surface_unpack_reconstruct", logging.info
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=functools.partial(
                surface_unpack_reconstruct,
                exploded_slice_size=args.exploded_slice_size,
            ),
        )



import typing

from ..frozen_instrumentation._HereditaryStratigraphicAssemblage import (
    HereditaryStratigraphicAssemblage,
)
from ._impl import col_records_from_pop_records
from ._specimen_from_records import specimen_from_records


def assemblage_from_records(
    records: typing.Dict,
    progress_wrap: typing.Callable = lambda x: x,
    mutate: bool = False,
) -> HereditaryStratigraphicAssemblage:
    """Deserialize a population of `HereditaryStratigraphicColumn`s into a
    `HereditaryStratigraphicAssemblage` from a dict composed of builtin
    types.

    Parameters
    ----------
    records : dict
        Data to deserialize.
    progress_wrap : Callable, optional
        Wrapper applied around generation iterator and row generator for final
        phylogeny compilation process.

        Pass tqdm or equivalent to display progress bars.
    mutate : bool, default False
        Are side effects on the input argument `records` allowed?

    See Also
    --------
    HereditaryStratigraphicAssemblage
        A collection of HereditaryStratigraphicSpecimens, padded to include
        entries for all ranks retained by any specimen within the assemblage.
    """

    col_records = col_records_from_pop_records(records, mutate=mutate)

    return HereditaryStratigraphicAssemblage(
        specimen_from_records(col_record)
        for col_record in progress_wrap(col_records)
    )



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcMrcaUncertaintyRelUpperBoundPessimalRank:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyRelUpperBoundPessimalRank",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: int,
        second_num_strata_deposited: int,
    ) -> int:
        """Calculate rank for which upper bound on relative MRCA uncertainty is
        pessimized."""
        return 0



import opytional as opyt
import pytest

from hstrat import hstrat
from hstrat.stratum_retention_strategy.stratum_retention_algorithms._impl import (
    CalcMrcaUncertaintyRelUpperBoundWorstCase,
)


@pytest.mark.parametrize(
    "policy_t",
    [
        hstrat.depth_proportional_resolution_algo.Policy,
        hstrat.depth_proportional_resolution_tapered_algo.Policy,
        hstrat.fixed_resolution_algo.Policy,
        # too slow, disable
        # hstrat.geom_seq_nth_root_algo.Policy,
        # hstrat.geom_seq_nth_root_tapered_algo.Policy,
        hstrat.recency_proportional_resolution_algo.Policy,
    ],
)
@pytest.mark.parametrize(
    "at_num_strata_deposited",
    [
        10**3,
        10**6,
    ],
)
@pytest.mark.parametrize(
    "at_rank",
    [
        # None, not yet implemented
        -2,
        -8,
        0,
        100,
    ],
)
@pytest.mark.parametrize(
    "target_value",
    [
        0.01,
        0.25,
        0.5,
    ],
)
def test_satisfiable_at_least(
    policy_t,
    at_num_strata_deposited,
    at_rank,
    target_value,
):

    if (
        CalcMrcaUncertaintyRelUpperBoundWorstCase()(
            None,
            at_num_strata_deposited,
            at_num_strata_deposited,
            opyt.or_value(at_rank, -1),
        )
        <= target_value
    ):
        # skip impossible to parameterize cases
        return

    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=target_value,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=at_num_strata_deposited,
            at_rank=at_rank,
        ),
        param_lower_bound=(
            0
            if policy_t == hstrat.recency_proportional_resolution_algo.Policy
            else 1
        ),
    )
    policy_spec = parameterizer(policy_t)
    assert policy_spec == policy_t(parameterizer=parameterizer).GetSpec()

    if at_rank is not None:
        assert (
            policy_t(policy_spec=policy_spec,).CalcMrcaUncertaintyRelExact(
                at_num_strata_deposited,
                at_num_strata_deposited,
                at_rank,
            )
            >= target_value
        )
    else:
        assert (
            policy_t(
                policy_spec=policy_spec,
            ).CalcMrcaUncertaintyRelExactAtPessimalRank(
                at_num_strata_deposited,
                at_num_strata_deposited,
            )
            >= target_value
        )

    if policy_t not in (
        hstrat.geom_seq_nth_root_algo.Policy,
        hstrat.geom_seq_nth_root_tapered_algo.Policy,
    ):
        # disable for these policies because too slow
        parameterizer = hstrat.PropertyAtLeastParameterizer(
            target_value=target_value,
            policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
                at_num_strata_deposited=at_num_strata_deposited,
                at_rank=at_rank,
            ),
            param_lower_bound=(
                0
                if policy_t
                == hstrat.recency_proportional_resolution_algo.Policy
                else 1
            ),
            param_upper_bound=None,
        )
        policy_spec = parameterizer(policy_t)
        assert policy_spec == policy_t(parameterizer=parameterizer).GetSpec()

        if at_rank is not None:
            assert (
                policy_t(policy_spec=policy_spec,).CalcMrcaUncertaintyRelExact(
                    at_num_strata_deposited,
                    at_num_strata_deposited,
                    at_rank,
                )
                >= target_value
            )
        else:
            assert (
                policy_t(
                    policy_spec=policy_spec,
                ).CalcMrcaUncertaintyRelExactAtPessimalRank(
                    at_num_strata_deposited,
                    at_num_strata_deposited,
                )
                >= target_value
            )


@pytest.mark.parametrize(
    "policy_t",
    [
        hstrat.depth_proportional_resolution_algo.Policy,
        hstrat.depth_proportional_resolution_tapered_algo.Policy,
        hstrat.fixed_resolution_algo.Policy,
        # too slow, disable
        # hstrat.geom_seq_nth_root_algo.Policy,
        # hstrat.geom_seq_nth_root_tapered_algo.Policy,
        hstrat.recency_proportional_resolution_algo.Policy,
    ],
)
@pytest.mark.parametrize(
    "at_num_strata_deposited",
    [
        10**3,
        10**6,
    ],
)
@pytest.mark.parametrize(
    "at_rank,target_value",
    [
        # (None, 10**9), not yet implemented
        (-1, 10**9),
        (-8, 10**9),
        (0, 1.01),
        (10, 1.1),
    ],
)
def test_unsatisfiable_at_least(
    policy_t,
    at_num_strata_deposited,
    at_rank,
    target_value,
):

    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=target_value,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=at_num_strata_deposited,
            at_rank=at_rank,
        ),
        param_lower_bound=1,
    )
    policy_spec = parameterizer(policy_t)

    assert policy_spec is None
    with pytest.raises(hstrat.UnsatisfiableParameterizationRequestError):
        policy_t(parameterizer=parameterizer)


@pytest.mark.parametrize(
    "policy_t",
    [
        hstrat.depth_proportional_resolution_algo.Policy,
        hstrat.depth_proportional_resolution_tapered_algo.Policy,
        hstrat.fixed_resolution_algo.Policy,
        # too slow, disable
        # hstrat.geom_seq_nth_root_algo.Policy,
        # hstrat.geom_seq_nth_root_tapered_algo.Policy,
        hstrat.recency_proportional_resolution_algo.Policy,
    ],
)
@pytest.mark.parametrize(
    "at_num_strata_deposited",
    [
        10**3,
        10**6,
    ],
)
@pytest.mark.parametrize(
    "at_rank",
    [
        # None, not yet implemented
        -1,
        -2,
        -8,
        100,
        0,
    ],
)
@pytest.mark.parametrize(
    "target_value",
    [
        0.01,
        0.25,
        1,
        1.1,
        100,
    ],
)
def test_satisfiable_at_most(
    policy_t,
    at_num_strata_deposited,
    at_rank,
    target_value,
):

    if (
        policy_t
        not in (
            hstrat.geom_seq_nth_root_algo.Policy,
            hstrat.geom_seq_nth_root_tapered_algo.Policy,
        )
        or target_value > 1
    ):
        parameterizer = hstrat.PropertyAtMostParameterizer(
            target_value=target_value,
            policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
                at_num_strata_deposited=at_num_strata_deposited,
                at_rank=at_rank,
            ),
            param_lower_bound=1,
            # without upper bound, test runs too slow
            param_upper_bound=10**6,
        )
        policy_spec = parameterizer(policy_t)
        assert policy_spec == policy_t(parameterizer=parameterizer).GetSpec()

        if at_rank is not None:
            assert (
                policy_t(policy_spec=policy_spec,).CalcMrcaUncertaintyRelExact(
                    at_num_strata_deposited,
                    at_num_strata_deposited,
                    at_rank,
                )
                <= target_value
            )
        else:
            assert (
                policy_t(
                    policy_spec=policy_spec,
                ).CalcMrcaUncertaintyRelExactAtPessimalRank(
                    at_num_strata_deposited,
                    at_num_strata_deposited,
                )
                <= target_value
            )

    if policy_t not in (
        hstrat.geom_seq_nth_root_algo.Policy,
        hstrat.geom_seq_nth_root_tapered_algo.Policy,
        hstrat.recency_proportional_resolution_algo.Policy,
    ):
        # don't evaluate geom seq nth root on these at all b/c too slow
        parameterizer = hstrat.PropertyAtMostParameterizer(
            target_value=target_value,
            policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
                at_num_strata_deposited=at_num_strata_deposited,
                at_rank=at_rank,
            ),
            param_lower_bound=1,
            param_upper_bound=None,
        )
        policy_spec = parameterizer(policy_t)
        assert policy_spec == policy_t(parameterizer=parameterizer).GetSpec()

        if at_rank is not None:
            assert (
                policy_t(policy_spec=policy_spec,).CalcMrcaUncertaintyRelExact(
                    at_num_strata_deposited,
                    at_num_strata_deposited,
                    at_rank,
                )
                <= target_value
            )
        else:
            assert (
                policy_t(
                    policy_spec=policy_spec,
                ).CalcMrcaUncertaintyRelExactAtPessimalRank(
                    at_num_strata_deposited,
                    at_num_strata_deposited,
                )
                <= target_value
            )


@pytest.mark.parametrize(
    "policy_t",
    [
        hstrat.depth_proportional_resolution_algo.Policy,
        hstrat.depth_proportional_resolution_tapered_algo.Policy,
        hstrat.fixed_resolution_algo.Policy,
        # too slow, disable
        # hstrat.geom_seq_nth_root_algo.Policy,
        # hstrat.geom_seq_nth_root_tapered_algo.Policy,
        hstrat.recency_proportional_resolution_algo.Policy,
    ],
)
@pytest.mark.parametrize(
    "at_num_strata_deposited",
    [
        10**3,
        10**6,
    ],
)
@pytest.mark.parametrize(
    "at_rank,target_value",
    [
        # (None, -1), not yet implemented
        (-1, -1),
        (-8, -1),
        (0, -1),
        (10, -1),
    ],
)
def test_unsatisfiable_at_most(
    policy_t,
    at_num_strata_deposited,
    at_rank,
    target_value,
):

    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=target_value,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=at_num_strata_deposited,
            at_rank=at_rank,
        ),
        param_lower_bound=1,
    )
    policy_spec = parameterizer(policy_t)

    assert policy_spec is None
    with pytest.raises(hstrat.UnsatisfiableParameterizationRequestError):
        policy_t(parameterizer=parameterizer)


def test_against_expected_upper_bound():

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=100,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=98,
        ),
        param_lower_bound=1,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=100,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=98,
        ),
        param_lower_bound=1,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.05,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.05,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.049,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.049,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=5,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.051,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=7,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.051,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )


def test_against_expected_no_upper_bound():

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=100,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=98,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=100,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.0,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=98,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    assert parameterizer(policy_t) == policy_t.policy_spec_t(
        fixed_resolution=1,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.05,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.05,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.049,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.049,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=5,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtLeastParameterizer(
        target_value=0.051,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=7,
    )

    # case
    policy_t = hstrat.fixed_resolution_algo.Policy
    parameterizer = hstrat.PropertyAtMostParameterizer(
        target_value=0.051,
        policy_evaluator=hstrat.MrcaUncertaintyRelExactEvaluator(
            at_num_strata_deposited=101,
            at_rank=0,
        ),
        param_lower_bound=1,
        param_upper_bound=None,
    )
    res_spec = parameterizer(policy_t)
    assert res_spec == policy_t.policy_spec_t(
        fixed_resolution=6,
    )



def calc_provided_resolution(
    size_curb: int,
    num_stratum_depositions_completed: int,
) -> int:
    """After n strata have been deposited, what resolution is provided?

    Resolution may be negative, indicating that geom seq nth root should be
    used.
    """
    # (resolution + 1) * log2(num_stratum_depositions_completed) <= size_curb
    # (resolution + 1) * ceil(log2(num_depositons)) == size_curb
    # resolution + 1 == size_curb / ceil(log2(num_depositons))
    # resolution = size_curb / ceil(log2(num_depositons)) - 1
    res = (
        size_curb
        // (
            # int cast allows for other integer-like types i.e., np int64 etc
            int(num_stratum_depositions_completed).bit_length()
            + 1
        )
        - 1
    )

    # ensure necessary ranks for consistent transition to
    # geometric sequence nth root algorithm
    # this threshold also used in calc_geom_seq_nth_root_transition_rank
    resolution_thresh = 2
    assert resolution_thresh >= 0

    if res < resolution_thresh:
        return -1
    else:
        return res



import numpy as np
import pandas as pd


def alifestd_has_contiguous_ids(phylogeny_df: pd.DataFrame) -> bool:
    """Do organisms ids' correspond to their row number?

    Input dataframe is not mutated by this operation.
    """
    return (
        phylogeny_df["id"].to_numpy() == np.arange(len(phylogeny_df))
    ).all()



from iterpop import iterpop as ip
import pandas as pd

from hstrat import hstrat


def test_hsa_empty():
    hsa = hstrat.HereditaryStratigraphicAssemblage([])
    assert [*hsa.BuildSpecimens()] == []


def test_hsa_singleton():
    spec1 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(data=[0, 8], index=[0, 1], dtype="uint8"), 8
    )
    hsa = hstrat.HereditaryStratigraphicAssemblage([spec1])
    assert all(
        ip.popsingleton(hsa.BuildSpecimens()).GetData() == spec1.GetData()
    )
    assert (
        ip.popsingleton(hsa.BuildSpecimens()).GetData().dtype.name == "UInt8"
    )
    assert (
        ip.popsingleton(hsa.BuildSpecimens()).GetStratumDifferentiaBitWidth()
        == spec1.GetStratumDifferentiaBitWidth()
    )
    assert 1 == len(
        set(
            specimen.GetStratumMask().shape
            for specimen in hsa.BuildSpecimens()
        )
    )
    assert 1 == len(
        set(
            specimen.GetDifferentiaVals().shape
            for specimen in hsa.BuildSpecimens()
        )
    )


def test_hsa_pair1():
    spec1 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(data=[42, 42], index=[0, 2], dtype="uint64"), 64
    )
    spec2 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(data=[111111, 0, 94], index=[0, 2, 3], dtype="uint64"), 64
    )
    hsa = hstrat.HereditaryStratigraphicAssemblage([spec1, spec2])
    assert 1 == len(
        set(
            specimen.GetStratumMask().shape
            for specimen in hsa.BuildSpecimens()
        )
    )
    assert 1 == len(
        set(
            specimen.GetDifferentiaVals().shape
            for specimen in hsa.BuildSpecimens()
        )
    )

    res1, res2 = hsa.BuildSpecimens()

    assert res1.GetData().equals(
        pd.Series(data=[42, 42, pd.NA], index=[0, 2, 3], dtype="UInt64")
    )
    assert (
        res1.GetStratumDifferentiaBitWidth()
        == spec1.GetStratumDifferentiaBitWidth()
    )
    assert res2.GetData().equals(
        pd.Series(data=[111111, 0, 94], index=[0, 2, 3], dtype="UInt64")
    )
    assert (
        res2.GetStratumDifferentiaBitWidth()
        == spec2.GetStratumDifferentiaBitWidth()
    )


def test_hsa_pair2():
    spec1 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(data=[42, 41, 42], index=[0, 1, 2], dtype="uint64"), 64
    )
    spec2 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(data=[111111, 0, 94], index=[0, 2, 3], dtype="uint64"), 64
    )
    hsa = hstrat.HereditaryStratigraphicAssemblage([spec1, spec2])
    assert 1 == len(
        set(
            specimen.GetStratumMask().shape
            for specimen in hsa.BuildSpecimens()
        )
    )
    assert 1 == len(
        set(
            specimen.GetDifferentiaVals().shape
            for specimen in hsa.BuildSpecimens()
        )
    )

    res1, res2 = hsa.BuildSpecimens()

    assert res1.GetData().equals(
        pd.Series(data=[42, 41, 42, pd.NA], index=[0, 1, 2, 3], dtype="UInt64")
    )
    assert (
        res1.GetStratumDifferentiaBitWidth()
        == spec1.GetStratumDifferentiaBitWidth()
    )
    assert res2.GetData().equals(
        pd.Series(
            data=[111111, pd.NA, 0, 94], index=[0, 1, 2, 3], dtype="UInt64"
        )
    )
    assert (
        res2.GetStratumDifferentiaBitWidth()
        == spec2.GetStratumDifferentiaBitWidth()
    )


def test_hsa_pair3():
    spec1 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(data=[42, 41, 42], index=[0, 1, 2], dtype="uint64"), 64
    )
    spec2 = hstrat.HereditaryStratigraphicSpecimen(
        pd.Series(
            data=[111111, 0, 94, 78], index=[0, 2, 3, 4], dtype="uint64"
        ),
        64,
    )
    hsa = hstrat.HereditaryStratigraphicAssemblage([spec1, spec2])
    assert 1 == len(
        set(
            specimen.GetStratumMask().shape
            for specimen in hsa.BuildSpecimens()
        )
    )
    assert 1 == len(
        set(
            specimen.GetDifferentiaVals().shape
            for specimen in hsa.BuildSpecimens()
        )
    )

    res1, res2 = hsa.BuildSpecimens()

    assert res1.GetData().equals(
        pd.Series(
            data=[42, 41, 42, pd.NA, pd.NA],
            index=[0, 1, 2, 3, 4],
            dtype="UInt64",
        )
    )
    assert (
        res1.GetStratumDifferentiaBitWidth()
        == spec1.GetStratumDifferentiaBitWidth()
    )
    assert res2.GetData().equals(
        pd.Series(
            data=[111111, pd.NA, 0, 94, 78],
            index=[0, 1, 2, 3, 4],
            dtype="UInt64",
        )
    )
    assert (
        res2.GetStratumDifferentiaBitWidth()
        == spec2.GetStratumDifferentiaBitWidth()
    )



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec
from .._impl import pick_policy


class CalcMrcaUncertaintyAbsUpperBoundPessimalRank:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(
        self: "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
        other: typing.Any,
    ) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcMrcaUncertaintyAbsUpperBoundPessimalRank",
        policy: PolicyCouplerBase,
        first_num_strata_deposited: typing.Optional[int],
        second_num_strata_deposited: typing.Optional[int],
    ) -> int:
        """Calculate rank for which upper bound on absolute MRCA uncertainty is
        pessimized."""
        return pick_policy(
            policy.GetSpec().GetSizeCurb(),
            max(first_num_strata_deposited, second_num_strata_deposited),
        ).CalcMrcaUncertaintyAbsUpperBoundPessimalRank(
            first_num_strata_deposited, second_num_strata_deposited
        )



import pandas as pd


def alifestd_make_empty(ancestor_id: bool = False) -> pd.DataFrame:
    """Create an alife standard phylogeny dataframe with zero rows."""
    res = pd.DataFrame(
        {
            "ancestor_list": pd.Series(dtype="str"),
            "id": pd.Series(dtype="int"),
        }
    )
    if ancestor_id:
        res["ancestor_id"] = pd.Series(dtype="int")
    return res



"""Classes to specify prior expectation for the distribution of MRCA generation
between hereditary stratigraphic columns, to assist in estimating the
phylogenetic relationship of hereditary stratigraphic columns."""

from ..._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import typing

import pandas as pd
import polars as pl

from ...._auxiliary_lib import anytree_iterative_deepcopy
from .._impl import TrieInnerNode
from ._detail import TriePostprocessorBase


class NopTriePostprocessor(TriePostprocessorBase):
    """Functor for nop trie postprocess."""

    def __call__(
        self: "NopTriePostprocessor",
        trie: typing.Union[TrieInnerNode, pd.DataFrame, pl.DataFrame],
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> typing.Union[TrieInnerNode, pd.DataFrame, pl.DataFrame]:
        """Apply postprocess functors to the input trie.

        Parameters
        ----------
        trie : TrieInnerNode or pd.DataFrame or pl.DataFrame
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.

            Not used in the current implementation.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?

            If False, a deep copy of the input trie is made.
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode or pd.DataFrame or pl.DataFrame
            The postprocessed trie, identical to input.
        """
        if isinstance(trie, TrieInnerNode):
            if not mutate:
                trie = anytree_iterative_deepcopy(
                    trie, progress_wrap=progress_wrap
                )
        elif isinstance(trie, pd.DataFrame):
            if not mutate:
                trie = trie.copy()
        elif isinstance(trie, pl.DataFrame):
            if not mutate:
                trie = trie.clone()
        else:
            raise TypeError
        return trie



import anytree

from hstrat import hstrat
import hstrat.phylogenetic_inference.tree._impl as impl
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.NopTriePostprocessor,
        detail.TriePostprocessorBase,
    )


def test_nop_mutate():
    root = impl.TrieInnerNode(rank=None, differentia=None, parent=None)
    inner = impl.TrieInnerNode(rank=0, differentia=0, parent=root)
    _leaf = impl.TrieLeafNode(parent=inner, taxon_label="A")
    root_ = hstrat.NopTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=False
    )
    assert root is not root_
    assert str(anytree.RenderTree(root).by_attr("id")) == str(
        anytree.RenderTree(root_).by_attr("id"),
    )

    root = hstrat.NopTriePostprocessor()(
        root, p_differentia_collision=0.5, mutate=True
    )
    assert str(anytree.RenderTree(root).by_attr("id")) == str(
        anytree.RenderTree(root_).by_attr("id"),
    )



import random
import string

import pytest

from hstrat._auxiliary_lib import (
    parse_from_numeral_system,
    render_to_numeral_system,
)


@pytest.fixture
def alphabet():
    return string.digits + string.ascii_lowercase


def test_render_to_numeral_system(alphabet):
    assert render_to_numeral_system(0, alphabet) == "0"
    assert render_to_numeral_system(1, alphabet) == "1"
    assert render_to_numeral_system(10, alphabet) == "a"
    assert render_to_numeral_system(35, alphabet) == "z"
    assert render_to_numeral_system(36, alphabet) == "10"
    assert render_to_numeral_system(100, alphabet) == "2s"
    assert render_to_numeral_system(123456, alphabet) == "2n9c"


def test_render_and_parse_numeral_system_consistency(alphabet):
    for __ in range(10):
        num = random.randint(0, 1000000)
        base_num = render_to_numeral_system(num, alphabet)
        assert parse_from_numeral_system(base_num, alphabet) == num

    for __ in range(10):
        num_digits = "".join([random.choice(alphabet) for j in range(10)])
        base_num = parse_from_numeral_system(num_digits, alphabet)
        assert render_to_numeral_system(base_num, alphabet) == num_digits


def test_builtin_conversion_operators_consistency():
    for __ in range(100):
        num = random.randint(0, 1000000)
        assert bin(num)[2:] == render_to_numeral_system(num, "01")
        assert hex(num)[2:] == render_to_numeral_system(
            num, "0123456789abcdef"
        )
        assert oct(num)[2:] == render_to_numeral_system(num, string.digits[:8])
        assert str(num) == render_to_numeral_system(num, string.digits)



# adapted from https://stackoverflow.com/a/11443871/17332200
def div_range(start: int, end: int, divide_by: int):

    assert divide_by > 0
    assert start >= 0
    assert end >= 0

    while start > end:
        yield start
        start //= divide_by



import argparse
import logging
import os
import pathlib
import typing

import more_itertools as mit
import numpy as np
import opytional as opyt
import pandas as pd
from tqdm import tqdm

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_mark_origin_time_delta_asexual import (
    alifestd_mark_origin_time_delta_asexual,
)
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_unfurl_traversal_postorder_asexual import (
    alifestd_unfurl_traversal_postorder_asexual,
)
from ._configure_prod_logging import configure_prod_logging
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration

# adapted from https://stackoverflow.com/a/3939381/17332200
_UNSAFE_SYMBOLS = ";(),[]:'"
_UNSAFE_TRANSLATION_TABLE = str.maketrans("", "", _UNSAFE_SYMBOLS)


def _format_newick_repr(taxon_label: str, origin_time_delta: str) -> str:
    # adapted from https://github.com/niemasd/TreeSwift/blob/63b8979fb5e616ba89079d44e594682683c1365e/treeswift/Node.py#L129
    label = taxon_label

    if label.translate(_UNSAFE_TRANSLATION_TABLE) != label:
        label = label.join("''")

    if origin_time_delta != "nan":
        if "." in origin_time_delta:
            origin_time_delta = origin_time_delta.rstrip("0").rstrip(".")
        label = f"{label}:{origin_time_delta}"

    return label


def _build_newick_string(
    ids: np.ndarray,
    labels: np.ndarray,
    origin_time_deltas: np.ndarray,
    ancestor_ids: np.ndarray,
    *,
    progress_wrap: typing.Callable,
) -> str:
    child_newick_reprs = dict()
    for id_, taxon_label, origin_time_delta, ancestor_id in progress_wrap(
        zip(ids, labels, origin_time_deltas, ancestor_ids)
    ):
        newick_repr = _format_newick_repr(taxon_label, origin_time_delta)

        children_reprs = child_newick_reprs.pop(id_, None)
        if children_reprs is not None:
            newick_repr = f"({','.join(children_reprs)}){newick_repr}"

        child_newick_reprs.setdefault(ancestor_id, []).append(newick_repr)

    logging.info(f"finalizing {len(child_newick_reprs)} subtrees...")
    return ";\n".join(map(mit.one, child_newick_reprs.values())) + ";"


def alifestd_as_newick_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
    *,
    label_key: typing.Optional[str] = None,
    progress_wrap=lambda x: x,
) -> str:
    """Convert phylogeny dataframe to Newick format.

    Parameters
    ----------
    phylogeny_df : pd.DataFrame
        Phylogeny dataframe in Alife standard format.
    mutate : bool, optional
        Allow in-place mutations of the input dataframe, by default False.
    label_key : str, optional
        Column to use for taxon labels, by default None.
    progress_wrap : typing.Callable, optional
        Pass tqdm or equivalent to display a progress bar.
    """

    logging.info(
        "creating newick string for alifestd df "
        f"with shape {phylogeny_df.shape}",
    )

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    logging.info("adding ancestor id column, if not present")
    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    logging.info("setting up `origin_time_delta` column...")
    if "origin_time_delta" in phylogeny_df.columns:
        logging.info("... already present!")
    elif "origin_time" in phylogeny_df.columns:
        logging.info("... calculating from `origin_time`...")
        phylogeny_df = alifestd_mark_origin_time_delta_asexual(
            phylogeny_df, mutate=True
        )
    else:
        logging.info("... marking null")
        phylogeny_df["origin_time_delta"] = np.nan

    logging.info("calculating postorder traversal order...")
    postorder_ids = alifestd_unfurl_traversal_postorder_asexual(phylogeny_df)

    logging.info("preparing labels...")
    phylogeny_df["__hstrat_label"] = opyt.apply_if_or_value(
        label_key, phylogeny_df.__getitem__, ""
    )
    phylogeny_df["__hstrat_label"] = phylogeny_df["__hstrat_label"].astype(str)

    logging.info("reshaping data...")
    reshaped = (
        phylogeny_df.loc[
            postorder_ids,
            ["id", "__hstrat_label", "origin_time_delta", "ancestor_id"],
        ]
        .astype({"origin_time_delta": str})
        .to_numpy()
        .T
    )

    logging.info("creating newick string...")
    result = _build_newick_string(*reshaped, progress_wrap=progress_wrap)

    logging.info(f"{len(result)=} {result[:20]=}")
    return result


_raw_description = """Convert Alife standard phylogeny data to Newick format.

Note that this CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--input-file",
        "-i",
        type=str,
        help="Alife standard dataframe file to convert to Newick format.",
    )
    parser.add_argument(
        "--output-file",
        "-o",
        type=str,
        help="Path to write Newick-formatted output to.",
    )
    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=get_hstrat_version(),
    )
    return parser


if __name__ == "__main__":
    configure_prod_logging()

    parser = _create_parser()
    args = parser.parse_args()
    input_ext = os.path.splitext(args.input_file)[1]

    logging.info(
        f"reading alife-standard {input_ext} phylogeny data from "
        f"{args.input_file}...",
    )
    phylogeny_df = {
        ".csv": pd.read_csv,
        ".fea": pd.read_feather,
        ".feather": pd.read_feather,
        ".pqt": pd.read_parquet,
        ".parquet": pd.read_parquet,
    }[input_ext](args.input_file)

    with log_context_duration(
        "hstrat._auxiliary_lib.alifestd_as_newick_asexual", logging.info
    ):
        logging.info("converting to Newick format...")
        newick_str = alifestd_as_newick_asexual(
            phylogeny_df, progress_wrap=tqdm
        )

    logging.info(f"writing Newick-formatted data to {args.output_file}...")
    pathlib.Path(args.output_file).write_text(newick_str)

    logging.info("done!")



import unittest

from hstrat._auxiliary_lib import alifestd_parse_ancestor_ids


class TestOmitLast(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_empty(self):
        assert alifestd_parse_ancestor_ids("[]") == []
        assert alifestd_parse_ancestor_ids("[None]") == []
        assert alifestd_parse_ancestor_ids("[none]") == []
        assert alifestd_parse_ancestor_ids("[NONE]") == []

    def test_singleton(self):
        assert alifestd_parse_ancestor_ids("[1]") == [1]
        assert alifestd_parse_ancestor_ids("[12]") == [12]
        assert alifestd_parse_ancestor_ids("[0]") == [0]

    def test_double(self):
        assert alifestd_parse_ancestor_ids("[1,12]") == [1, 12]
        assert alifestd_parse_ancestor_ids("[12,1]") == [12, 1]
        assert alifestd_parse_ancestor_ids("[0,1]") == [0, 1]

        assert alifestd_parse_ancestor_ids("[1, 0]") == [1, 0]
        assert alifestd_parse_ancestor_ids("[12, 1]") == [12, 1]
        assert alifestd_parse_ancestor_ids("[0, 1]") == [0, 1]

    def test_triple(self):
        assert alifestd_parse_ancestor_ids("[1,12,128]") == [1, 12, 128]
        assert alifestd_parse_ancestor_ids("[12,1, 0]") == [12, 1, 0]
        assert alifestd_parse_ancestor_ids("[0, 1,143]") == [0, 1, 143]
        assert alifestd_parse_ancestor_ids("[1, 0, 199]") == [1, 0, 199]


if __name__ == "__main__":
    unittest.main()



import itertools as it
import typing


def with_omission(
    iterable: typing.Iterable, omit_index: int
) -> typing.Generator:
    """Return an iterator that yields the elements of `iterable`, skipping over
    the element at index `omit_index`.

    Yields
    ------
    object
        The next element in the iterable with the specified index omitted.

    Examples
    --------
    >>> list(with_omission([1, 2, 3], 0))
    [2, 3]

    >>> list(with_omission('abcde', 2))
    ['a', 'b', 'd', 'e']

    >>> list(with_omission((1, 2, 3), 2))
    [1, 2]
    """

    iterator = iter(iterable)
    yield from it.islice(iterator, omit_index)
    next(iterator, None)
    yield from iterator



from ._PropertyAtLeastParameterizer import PropertyAtLeastParameterizer
from ._PropertyAtMostParameterizer import PropertyAtMostParameterizer
from ._PropertyExactlyParameterizer import PropertyExactlyParameterizer

# adapted from https://stackoverflow.com/a/31079085
__all__ = [
    "PropertyAtLeastParameterizer",
    "PropertyAtMostParameterizer",
    "PropertyExactlyParameterizer",
]



from ._AnyTreeAscendingIter import AnyTreeAscendingIter
from ._AnyTreeFastLeafIter import AnyTreeFastLeafIter
from ._AnyTreeFastLevelOrderIter import AnyTreeFastLevelOrderIter
from ._AnyTreeFastPostOrderIter import AnyTreeFastPostOrderIter
from ._AnyTreeFastPreOrderIter import AnyTreeFastPreOrderIter
from ._BioPhyloTree import BioPhyloTree
from ._CopyableSeriesItemsIter import CopyableSeriesItemsIter
from ._GetAttrLaunderShim import GetAttrLaunderShim
from ._HereditaryStratigraphicArtifact import HereditaryStratigraphicArtifact
from ._RecursionLimit import RecursionLimit
from ._RngStateContext import RngStateContext
from ._ScalarFormatterFixedPrecision import ScalarFormatterFixedPrecision
from ._alifestd_aggregate_phylogenies import alifestd_aggregate_phylogenies
from ._alifestd_as_newick_asexual import alifestd_as_newick_asexual
from ._alifestd_assign_contiguous_ids import alifestd_assign_contiguous_ids
from ._alifestd_assign_root_ancestor_token import (
    alifestd_assign_root_ancestor_token,
)
from ._alifestd_calc_polytomic_index import alifestd_calc_polytomic_index
from ._alifestd_categorize_triplet_asexual import (
    alifestd_categorize_triplet_asexual,
)
from ._alifestd_chronological_sort import alifestd_chronological_sort
from ._alifestd_coarsen_mask import alifestd_coarsen_mask
from ._alifestd_coerce_chronological_consistency import (
    alifestd_coerce_chronological_consistency,
)
from ._alifestd_collapse_trunk_asexual import alifestd_collapse_trunk_asexual
from ._alifestd_collapse_unifurcations import alifestd_collapse_unifurcations
from ._alifestd_convert_root_ancestor_token import (
    alifestd_convert_root_ancestor_token,
)
from ._alifestd_count_children_of_asexual import (
    alifestd_count_children_of_asexual,
)
from ._alifestd_count_inner_nodes import alifestd_count_inner_nodes
from ._alifestd_count_leaf_nodes import alifestd_count_leaf_nodes
from ._alifestd_count_polytomies import alifestd_count_polytomies
from ._alifestd_count_root_nodes import alifestd_count_root_nodes
from ._alifestd_count_unifurcations import alifestd_count_unifurcations
from ._alifestd_delete_trunk_asexual import alifestd_delete_trunk_asexual
from ._alifestd_downsample_tips_asexual import alifestd_downsample_tips_asexual
from ._alifestd_estimate_triplet_distance_asexual import (
    alifestd_estimate_triplet_distance_asexual,
)
from ._alifestd_find_chronological_inconsistency import (
    alifestd_find_chronological_inconsistency,
)
from ._alifestd_find_leaf_ids import alifestd_find_leaf_ids
from ._alifestd_find_mrca_id_asexual import alifestd_find_mrca_id_asexual
from ._alifestd_find_root_ids import alifestd_find_root_ids
from ._alifestd_has_compact_ids import alifestd_has_compact_ids
from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_has_increasing_ids import alifestd_has_increasing_ids
from ._alifestd_has_multiple_roots import alifestd_has_multiple_roots
from ._alifestd_is_asexual import alifestd_is_asexual
from ._alifestd_is_chronologically_ordered import (
    alifestd_is_chronologically_ordered,
)
from ._alifestd_is_chronologically_sorted import (
    alifestd_is_chronologically_sorted,
)
from ._alifestd_is_sexual import alifestd_is_sexual
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_join_roots import alifestd_join_roots
from ._alifestd_make_ancestor_id_col import alifestd_make_ancestor_id_col
from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_make_empty import alifestd_make_empty
from ._alifestd_mark_ancestor_origin_time_asexual import (
    alifestd_mark_ancestor_origin_time_asexual,
)
from ._alifestd_mark_leaves import alifestd_mark_leaves
from ._alifestd_mark_node_depth_asexual import alifestd_mark_node_depth_asexual
from ._alifestd_mark_num_children_asexual import (
    alifestd_mark_num_children_asexual,
)
from ._alifestd_mark_num_descendants_asexual import (
    alifestd_mark_num_descendants_asexual,
)
from ._alifestd_mark_num_leaves_asexual import alifestd_mark_num_leaves_asexual
from ._alifestd_mark_oldest_root import alifestd_mark_oldest_root
from ._alifestd_mark_origin_time_delta_asexual import (
    alifestd_mark_origin_time_delta_asexual,
)
from ._alifestd_mark_ot_mrca_asexual import alifestd_mark_ot_mrca_asexual
from ._alifestd_mark_root_id import alifestd_mark_root_id
from ._alifestd_mark_roots import alifestd_mark_roots
from ._alifestd_parse_ancestor_id import alifestd_parse_ancestor_id
from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids
from ._alifestd_prune_extinct_lineages_asexual import (
    alifestd_prune_extinct_lineages_asexual,
)
from ._alifestd_reroot_at_id_asexual import alifestd_reroot_at_id_asexual
from ._alifestd_sample_triplet_comparisons_asexual import (
    alifestd_sample_triplet_comparisons_asexual,
)
from ._alifestd_splay_polytomies import alifestd_splay_polytomies
from ._alifestd_sum_origin_time_deltas_asexual import (
    alifestd_sum_origin_time_deltas_asexual,
)
from ._alifestd_to_working_format import alifestd_to_working_format
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col
from ._alifestd_try_add_ancestor_list_col import (
    alifestd_try_add_ancestor_list_col,
)
from ._alifestd_unfurl_lineage_asexual import alifestd_unfurl_lineage_asexual
from ._alifestd_unfurl_traversal_postorder_asexual import (
    alifestd_unfurl_traversal_postorder_asexual,
)
from ._alifestd_validate import alifestd_validate
from ._all_same import all_same
from ._all_unique import all_unique
from ._anynode_deepcopy_except_neighbors import (
    anynode_deepcopy_except_neighbors,
)
from ._anytree_calc_leaf_counts import anytree_calc_leaf_counts
from ._anytree_cardinality import anytree_cardinality
from ._anytree_has_grandparent import anytree_has_grandparent
from ._anytree_has_sibling import anytree_has_sibling
from ._anytree_iterative_deepcopy import anytree_iterative_deepcopy
from ._anytree_peel_sibling_to_cousin import anytree_peel_sibling_to_cousin
from ._anytree_tree_to_alife_dataframe import anytree_tree_to_alife_dataframe
from ._apply_swaps import apply_swaps
from ._argsort import argsort
from ._as_compact_type import as_compact_type
from ._as_nullable_type import as_nullable_type
from ._assign_intersecting_subsets import assign_intersecting_subsets
from ._bit_ceil import bit_ceil
from ._bit_drop_msb import bit_drop_msb
from ._bit_floor import bit_floor
from ._bytes_swap_bit_order import bytes_swap_bit_order
from ._capitalize_n import capitalize_n
from ._caretdown_marker import caretdown_marker
from ._caretup_marker import caretup_marker
from ._cast_int_lossless import cast_int_lossless
from ._check_testing_requirements import check_testing_requirements
from ._cmp import cmp
from ._cmp_approx import cmp_approx
from ._coerce_to_pandas import coerce_to_pandas
from ._coerce_to_polars import coerce_to_polars
from ._collapse_nonleading_whitespace import collapse_nonleading_whitespace
from ._configure_prod_logging import configure_prod_logging
from ._consume import consume
from ._coshuffled import coshuffled
from ._count_leading_blanks import count_leading_blanks
from ._count_trailing_ones import count_trailing_ones
from ._count_trailing_zeros import count_trailing_zeros
from ._count_unique import count_unique
from ._curried_binary_search_jit import curried_binary_search_jit
from ._deep_listify import deep_listify
from ._delegate_polars_implementation import delegate_polars_implementation
from ._demark import demark
from ._div_range import div_range
from ._estimate_binomial_p import estimate_binomial_p
from ._except_wrap_sentinel import except_wrap_sentinel
from ._find_bounds import find_bounds
from ._flag_last import flag_last
from ._flat_len import flat_len
from ._format_cli_description import format_cli_description
from ._generate_n import generate_n
from ._generate_omission_subsets import generate_omission_subsets
from ._get_hstrat_version import get_hstrat_version
from ._get_nullable_mask import get_nullable_mask
from ._get_nullable_vals import get_nullable_vals
from ._get_package_name import get_package_name
from ._get_sole_scalar_value_polars import get_sole_scalar_value_polars
from ._give_len import give_len
from ._indices_of_unique import indices_of_unique
from ._intersect_ranges import intersect_ranges
from ._is_base64 import is_base64
from ._is_in import is_in
from ._is_in_coverage_run import is_in_coverage_run
from ._is_in_unit_test import is_in_unit_test
from ._is_nondecreasing import is_nondecreasing
from ._is_nonincreasing import is_nonincreasing
from ._is_strictly_decreasing import is_strictly_decreasing
from ._is_strictly_increasing import is_strictly_increasing
from ._is_subset import is_subset
from ._iter_chunks import iter_chunks
from ._iter_monotonic_equivalencies import iter_monotonic_equivalencies
from ._iter_monotonic_equivalencies_reverse import (
    iter_monotonic_equivalencies_reverse,
)
from ._iter_slices import iter_slices
from ._jit import jit
from ._jit_TypingError import jit_TypingError
from ._jit_numba_dict_t import jit_numba_dict_t
from ._jit_numba_integer_array_ts import (
    jit_numba_int8_arr_t,
    jit_numba_int16_arr_t,
    jit_numba_int32_arr_t,
    jit_numba_int64_arr_t,
    jit_numba_uint8_arr_t,
    jit_numba_uint16_arr_t,
    jit_numba_uint32_arr_t,
    jit_numba_uint64_arr_t,
)
from ._jit_numpy_bool_t import jit_numpy_bool_t
from ._jit_numpy_int64_t import jit_numpy_int64_t
from ._join_paragraphs_from_one_sentence_per_line import (
    join_paragraphs_from_one_sentence_per_line,
)
from ._launder_impl_modules import launder_impl_modules
from ._lazy_attach import lazy_attach
from ._lazy_attach_stub import lazy_attach_stub
from ._load_cppimportable_module import load_cppimportable_module
from ._log_context_duration import log_context_duration
from ._log_memory_usage import log_memory_usage
from ._log_once_in_a_row import log_once_in_a_row
from ._make_intersecting_subsets import make_intersecting_subsets
from ._memoize_generator import memoize_generator
from ._min_array_dtype import min_array_dtype
from ._numpy_fromiter_polyfill import numpy_fromiter_polyfill
from ._numpy_index import numpy_index
from ._numpy_index_flat import numpy_index_flat
from ._omit_last import omit_last
from ._pairwise import pairwise
from ._parse_from_numeral_system import parse_from_numeral_system
from ._popcount import popcount
from ._raises import raises
from ._random_choice_generator import random_choice_generator
from ._random_tree import random_tree
from ._release_cur_mpl_fig import release_cur_mpl_fig
from ._render_pandas_snapshot import render_pandas_snapshot
from ._render_polars_snapshot import render_polars_snapshot
from ._render_to_base64url import render_to_base64url
from ._render_to_numeral_system import render_to_numeral_system
from ._reversed_enumerate import reversed_enumerate
from ._reversed_range import reversed_range
from ._scale_luminosity import scale_luminosity
from ._seed_random import seed_random
from ._splicewhile import splicewhile
from ._swap_rows_and_indices import swap_rows_and_indices
from ._textwrap_respect_indents import textwrap_respect_indents
from ._to_tril import to_tril
from ._unfurl_lineage_with_contiguous_ids import (
    unfurl_lineage_with_contiguous_ids,
)
from ._unpairwise import unpairwise
from ._unzip import unzip
from ._warn_once import warn_once
from ._with_omission import with_omission
from ._with_rng_state_context import with_rng_state_context
from ._zip_strict import zip_strict

__all__ = [
    "alifestd_aggregate_phylogenies",
    "alifestd_as_newick_asexual",
    "alifestd_assign_contiguous_ids",
    "alifestd_assign_root_ancestor_token",
    "alifestd_categorize_triplet_asexual",
    "alifestd_chronological_sort",
    "alifestd_coarsen_mask",
    "alifestd_coerce_chronological_consistency",
    "alifestd_collapse_trunk_asexual",
    "alifestd_collapse_unifurcations",
    "alifestd_calc_polytomic_index",
    "alifestd_count_children_of_asexual",
    "alifestd_count_inner_nodes",
    "alifestd_count_leaf_nodes",
    "alifestd_count_root_nodes",
    "alifestd_count_polytomies",
    "alifestd_count_unifurcations",
    "alifestd_delete_trunk_asexual",
    "alifestd_downsample_tips_asexual",
    "alifestd_estimate_triplet_distance_asexual",
    "alifestd_find_chronological_inconsistency",
    "alifestd_find_leaf_ids",
    "alifestd_find_mrca_id_asexual",
    "alifestd_find_root_ids",
    "alifestd_has_compact_ids",
    "alifestd_has_contiguous_ids",
    "alifestd_has_increasing_ids",
    "alifestd_has_multiple_roots",
    "alifestd_is_asexual",
    "alifestd_is_chronologically_ordered",
    "alifestd_is_chronologically_sorted",
    "alifestd_is_sexual",
    "alifestd_is_topologically_sorted",
    "alifestd_join_roots",
    "alifestd_make_ancestor_id_col",
    "alifestd_make_ancestor_list_col",
    "alifestd_make_empty",
    "alifestd_mark_ancestor_origin_time_asexual",
    "alifestd_mark_leaves",
    "alifestd_mark_node_depth_asexual",
    "alifestd_mark_num_children_asexual",
    "alifestd_mark_num_descendants_asexual",
    "alifestd_mark_num_leaves_asexual",
    "alifestd_mark_oldest_root",
    "alifestd_mark_origin_time_delta_asexual",
    "alifestd_mark_ot_mrca_asexual",
    "alifestd_mark_root_id",
    "alifestd_mark_roots",
    "alifestd_parse_ancestor_id",
    "alifestd_parse_ancestor_ids",
    "alifestd_prune_extinct_lineages_asexual",
    "alifestd_reroot_at_id_asexual",
    "alifestd_convert_root_ancestor_token",
    "alifestd_sample_triplet_comparisons_asexual",
    "alifestd_splay_polytomies",
    "alifestd_sum_origin_time_deltas_asexual",
    "alifestd_to_working_format",
    "alifestd_topological_sort",
    "alifestd_try_add_ancestor_id_col",
    "alifestd_try_add_ancestor_list_col",
    "alifestd_unfurl_lineage_asexual",
    "alifestd_unfurl_traversal_postorder_asexual",
    "alifestd_validate",
    "all_same",
    "all_unique",
    "anynode_deepcopy_except_neighbors",
    "anytree_calc_leaf_counts",
    "anytree_cardinality",
    "anytree_iterative_deepcopy",
    "anytree_has_grandparent",
    "anytree_has_sibling",
    "anytree_peel_sibling_to_cousin",
    "anytree_tree_to_alife_dataframe",
    "apply_swaps",
    "argsort",
    "as_compact_type",
    "as_nullable_type",
    "assign_intersecting_subsets",
    "AnyTreeAscendingIter",
    "AnyTreeFastLeafIter",
    "AnyTreeFastLevelOrderIter",
    "AnyTreeFastPreOrderIter",
    "AnyTreeFastPostOrderIter",
    "BioPhyloTree",
    "bit_ceil",
    "bit_drop_msb",
    "bit_floor",
    "bytes_swap_bit_order",
    "capitalize_n",
    "caretdown_marker",
    "caretup_marker",
    "cast_int_lossless",
    "check_testing_requirements",
    "cmp",
    "cmp_approx",
    "coerce_to_pandas",
    "coerce_to_polars",
    "collapse_nonleading_whitespace",
    "configure_prod_logging",
    "consume",
    "CopyableSeriesItemsIter",
    "count_unique",
    "coshuffled",
    "count_leading_blanks",
    "count_trailing_ones",
    "count_trailing_zeros",
    "curried_binary_search_jit",
    "deep_listify",
    "delegate_polars_implementation",
    "demark",
    "div_range",
    "estimate_binomial_p",
    "except_wrap_sentinel",
    "find_bounds",
    "flag_last",
    "flat_len",
    "format_cli_description",
    "generate_n",
    "generate_omission_subsets",
    "get_hstrat_version",
    "get_nullable_mask",
    "get_nullable_vals",
    "get_package_name",
    "get_sole_scalar_value_polars",
    "GetAttrLaunderShim",
    "give_len",
    "HereditaryStratigraphicArtifact",
    "indices_of_unique",
    "intersect_ranges",
    "is_base64",
    "is_in",
    "is_in_coverage_run",
    "is_in_unit_test",
    "is_nondecreasing",
    "is_nonincreasing",
    "is_strictly_decreasing",
    "is_strictly_increasing",
    "is_subset",
    "iter_chunks",
    "iter_monotonic_equivalencies",
    "iter_monotonic_equivalencies_reverse",
    "iter_slices",
    "jit",
    "jit_numba_int8_arr_t",
    "jit_numba_int16_arr_t",
    "jit_numba_int32_arr_t",
    "jit_numba_int64_arr_t",
    "jit_numba_uint8_arr_t",
    "jit_numba_uint16_arr_t",
    "jit_numba_uint32_arr_t",
    "jit_numba_uint64_arr_t",
    "jit_numba_dict_t",
    "jit_numpy_bool_t",
    "jit_numpy_int64_t",
    "jit_TypingError",
    "join_paragraphs_from_one_sentence_per_line",
    "launder_impl_modules",
    "lazy_attach",
    "lazy_attach_stub",
    "load_cppimportable_module",
    "log_context_duration",
    "log_memory_usage",
    "log_once_in_a_row",
    "make_intersecting_subsets",
    "memoize_generator",
    "min_array_dtype",
    "numpy_fromiter_polyfill",
    "numpy_index",
    "numpy_index_flat",
    "omit_last",
    "pairwise",
    "parse_from_numeral_system",
    "popcount",
    "RecursionLimit",
    "raises",
    "release_cur_mpl_fig",
    "render_pandas_snapshot",
    "render_polars_snapshot",
    "random_choice_generator",
    "random_tree",
    "render_to_base64url",
    "render_to_numeral_system",
    "reversed_enumerate",
    "reversed_range",
    "RngStateContext",
    "scale_luminosity",
    "ScalarFormatterFixedPrecision",
    "seed_random",
    "splicewhile",
    "swap_rows_and_indices",
    "textwrap_respect_indents",
    "to_tril",
    "unfurl_lineage_with_contiguous_ids",
    "unpairwise",
    "unzip",
    "warn_once",
    "with_omission",
    "with_rng_state_context",
    "zip_strict",
]
